#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to check for fully specified environment."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING
from itertools import chain
from packaging.version import Version
from packaging.specifiers import SpecifierSet

import attr
from thoth.common import get_justification_link as jl

from ..boot import Boot

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class VersionCheckBoot(Boot):
    """A boot that checks if versions are too lax."""

    _JUSTIFICATION_VERSION_TOO_LAX = jl("lax_version")

    # Some arbitrary large version to detect possibly version specifiers without any upper limit.
    _LARGE_VERSION = Version("9999999999")

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self, always."""
        if builder_context.is_adviser_pipeline() and not builder_context.is_included(cls):
            yield {}
            return None

        yield from ()
        return None

    def run(self) -> None:
        """Check if versions are not too lax."""
        for package_version in chain(
            self.context.project.pipfile.packages.packages.values(),
            self.context.project.pipfile.dev_packages.packages.values(),
        ):
            if not package_version.version or package_version.version == "*":
                self.context.stack_info.append(
                    {
                        "type": "WARNING",
                        "message": f"No version range specifier for {package_version.name!r} found, it is "
                        f"recommended to specify version ranges in requirements",
                        "link": self._JUSTIFICATION_VERSION_TOO_LAX,
                    }
                )
            elif self._LARGE_VERSION in SpecifierSet(package_version.version):
                self.context.stack_info.append(
                    {
                        "type": "WARNING",
                        "message": f"Version range specifier ({package_version.version!r}) for "
                        f"{package_version.name!r} might be too lax",
                        "link": self._JUSTIFICATION_VERSION_TOO_LAX,
                    }
                )



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Penalize stacks with a CVE."""

import logging

from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion
from thoth.storages.exceptions import NotFoundError
from voluptuous import Required
from voluptuous import Schema

from ..enums import RecommendationType
from ..step import Step
from ..state import State

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class CvePenalizationStep(Step):
    """Penalization based on CVE being present in stack."""

    CONFIGURATION_DEFAULT = {"package_name": None, "cve_penalization": -0.2, "multi_package_resolution": False}
    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("package_name"): None,
            Required("cve_penalization"): float,
            Required("multi_package_resolution"): False,
        }
    )
    _JUSTIFICATION_LINK = jl("cve")
    _JUSTIFICATION_LINK_NO_CVE = jl("no_cve")

    _messages_logged = attr.ib(type=Set[Tuple[str, str, str]], factory=set, init=False)
    _allow_cves = attr.ib(type=Set[str], factory=set, init=False)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Remove CVEs only for advised stacks."""
        if (
            builder_context.is_adviser_pipeline()
            and not builder_context.is_included(cls)
            and builder_context.recommendation_type != RecommendationType.SECURITY
        ):
            if builder_context.recommendation_type in (RecommendationType.LATEST, RecommendationType.TESTING):
                # Report no cve penalization for latest and testing recommendation types.
                yield {"cve_penalization": 0.0}
            else:
                yield {}

            return None

        yield from ()
        return None

    def pre_run(self) -> None:
        """Initialize this pipeline unit before running."""
        self._messages_logged.clear()
        self._construct_allow_cves(self._allow_cves, self.context.labels)
        super().pre_run()

    def run(self, _: State, package_version: PackageVersion) -> Optional[Tuple[float, List[Dict[str, str]]]]:
        """Penalize stacks with a CVE."""
        try:
            cve_records = self.context.graph.get_python_cve_records_all(
                package_name=package_version.name,
                package_version=package_version.locked_version,
            )
        except NotFoundError as exc:
            _LOGGER.warning("Package %r in version %r not found: %r", exc)
            return None

        if cve_records:
            package_version_tuple = package_version.to_tuple()
            _LOGGER.debug("Found CVEs for %r: %r", package_version_tuple, cve_records)

            justification = []
            for cve_record in cve_records:
                message = f"Package  {package_version_tuple!r} has a CVE {cve_record['cve_id']!r}"
                justification.append(
                    {
                        "package_name": package_version.name,
                        "link": cve_record.get("link") or self._JUSTIFICATION_LINK,
                        "advisory": cve_record["details"],
                        "message": message,
                        "type": "WARNING",
                    }
                )

            if self.context.recommendation_type not in (RecommendationType.LATEST, RecommendationType.TESTING):
                # Penalize only if not latest/testing and based on `allow-cve' label supplied.
                penalized_cves = [
                    cve_record for cve_record in cve_records if cve_record["cve_id"] not in self._allow_cves
                ]
                penalization = len(penalized_cves) * self.configuration["cve_penalization"]
                return max(penalization, -1.0), justification

            return 0.0, justification
        else:
            justification = [
                {
                    "package_name": package_version.name,
                    "link": self._JUSTIFICATION_LINK_NO_CVE,
                    "type": "INFO",
                    "message": f"No known CVE known for {package_version.name!r} in "
                    f"version {package_version.locked_version!r}",
                }
            ]
            return 0.0, justification



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Database of CPUs and CPU flags."""

import os
import yaml

from typing import List


class _CPUDatabase:
    """A database of CPUs and CPU flags."""

    _CPU_DB_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "cpu_db.yaml")

    __slots__ = ["_cpu_flags"]

    def __init__(self) -> None:
        """Load and prepare the database."""
        self._cpu_flags = {}

        with open(self._CPU_DB_PATH, "r") as cpu_db:
            content = yaml.load(cpu_db, Loader=yaml.CLoader)

        for flag, cpus in content["flags"].items():
            self._cpu_flags[flag] = frozenset(tuple(i) for i in cpus)

    def provides_flag(self, cpu_model: int, cpu_family: int, flag: str) -> bool:
        """Check if the given CPU provides the given flag."""
        return (cpu_model, cpu_family) in (self._cpu_flags.get(flag) or frozenset())

    def get_known_flags(self) -> List[str]:
        """Get known CPU flags."""
        return list(self._cpu_flags.keys())


CPUDatabase = _CPUDatabase()



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation of Beam for beam searching part of adviser."""

import random
from typing import Any
from typing import List
from typing import Tuple
from typing import Generator
from typing import Optional
from typing import TYPE_CHECKING
import logging

from fext import ExtHeapQueue

import attr

from .exceptions import NoHistoryKept
from .state import State
from .utils import should_keep_history

if TYPE_CHECKING:
    import matplotlib

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class Beam:
    """Beam implementation.

    The implementation of beam respecting beam width: https://en.wikipedia.org/wiki/Beam_search

    Beam is internally implemented on top of a min-heap queue to optimize inserts and respect
    beam width in O(log(N)).

    The most frequent operation performed on the beam is not always the same - it depends
    on the beam width and number of states generated. If number of states generated  is
    more than the width of the beam, its reasonable to optimize insertions into the beam
    with checks on beam width.

    If number of states generated is smaller than the beam width, the beam could be optimized
    for removal of states.

    As the first scenario is seen in a real-world deployment, the beam uses min-heapq to keep
    addition to the beam with beam_width checks in O(log(N)) and removals of the states in
    O(log(N)). To satisfy removals in O(log(N)), the beam maintains a dictionary mapping a state
    to its index in the beam.
    """

    width = attr.ib(default=None, type=Optional[int])
    keep_history = attr.ib(type=bool, kw_only=True, default=None, converter=should_keep_history)

    _beam_history = attr.ib(type=List[Tuple[int, Optional[float]]], default=attr.Factory(list), kw_only=True)

    _heap = attr.ib(type=ExtHeapQueue, init=False)
    _WIDTH_VALIDATOR_ERR_MSG = "Beam width has to be None or positive integer, got {!r}"

    @width.validator
    def _validate_width(self, attribute: Any, value: Optional[int]) -> None:
        """Validate width initialization."""
        assert attribute.name == "width", "Wrong attribute to validate"

        if value is None:
            return

        if isinstance(value, int):
            if value <= 0:
                raise ValueError(self._WIDTH_VALIDATOR_ERR_MSG.format(value))

            return

        raise ValueError(self._WIDTH_VALIDATOR_ERR_MSG.format(value))

    @_heap.default
    def _heap_default(self) -> ExtHeapQueue:
        """Initialize the extended heap queue."""
        if self.width is not None:
            return ExtHeapQueue(size=self.width)

        return ExtHeapQueue()

    def new_iteration(self) -> None:  # noqa: D401
        """Called once a new iteration is done in resolver.

        Used to keep track of beam history and to keep track of states added.
        """
        if not self.keep_history:
            return

        self._beam_history.append((self.size, self.max().score if self.size > 0 else None))

    @staticmethod
    def _make_patch_spines_invisible(ax: Any) -> None:
        """Make spines invisible."""
        ax.set_frame_on(True)
        ax.patch.set_visible(False)
        for sp in ax.spines.values():
            sp.set_visible(False)

    def plot(self) -> "matplotlib.figure.Figure":
        """Plot temperature history of adaptive simulated annealing."""
        if not self._beam_history:
            raise NoHistoryKept("No history datapoints kept for beam")

        import matplotlib.pyplot as plt
        from matplotlib.font_manager import FontProperties

        x = [i for i in range(len(self._beam_history))]
        # Beam size over time.
        y1 = [i[0] for i in self._beam_history]
        # Highest rated state history.
        y2 = [i[1] for i in self._beam_history]

        fig, host = plt.subplots()
        fig.subplots_adjust(right=0.75)

        par1 = host.twinx()

        par1.spines["right"].set_position(("axes", 1.10))
        self._make_patch_spines_invisible(par1)

        par1.spines["right"].set_visible(True)
        host.spines["right"].set_visible(False)
        host.spines["top"].set_visible(False)

        (p1,) = host.plot(x, y1, ",g", label="Beam size")
        (p2,) = par1.plot(x, y2, ",b", label="Score of top rated partially resolved state")

        host.set_xlabel("iteration")
        host.set_ylabel("beam size")
        par1.set_ylabel("score")

        host.yaxis.label.set_color(p1.get_color())
        par1.yaxis.label.set_color(p2.get_color())

        tkw = dict(size=4, width=1.5)
        host.tick_params(axis="y", colors=p1.get_color(), **tkw)
        host.tick_params(axis="x", **tkw)
        par1.tick_params(axis="y", colors=p2.get_color(), **tkw)

        font_prop = FontProperties()
        font_prop.set_size("medium")
        fig.legend(
            loc="upper center",
            bbox_to_anchor=(0.50, 1.00),
            ncol=2,
            fancybox=True,
            shadow=True,
            prop=font_prop,
        )
        return fig

    @property
    def size(self) -> int:
        """Get the current size of beam."""
        return len(self._heap)

    def wipe(self) -> None:
        """Remove all states from beam."""
        self._beam_history.clear()
        self._heap.clear()

    def iter_states(self) -> List[State]:
        """Iterate over states, do not respect their score in order of iteration."""
        to_return = self._heap.items()  # type: List[State]
        return to_return

    def iter_states_sorted(self, reverse: bool = True) -> Generator[State, None, None]:
        """Iterate over sorted states."""
        return (item for item in sorted(self._heap.items(), reverse=reverse))

    def max(self) -> State:
        """Return the highest rated state as kept in the beam."""
        to_return = self._heap.get_max()  # type: State
        return to_return

    def add_state(self, state: State) -> None:
        """Add state to the internal state listing (do it in O(log(N)) time."""
        self._heap.push(state.score, state)

    def get(self, idx: int) -> State:
        """Get i-th element from the beam (constant time), keep it in the beam.

        This method is suitable for random state indexing in the beam (like in case of adaptive
        simulated annealing). The actual index is not into a sorted array and has no special meaning
        assigned - beam under the hood uses min-heapq (as of now), but the index used is not guaranteed to
        point to a heap-like data structure.
        """
        to_return = self._heap.get(idx)  # type: State
        return to_return

    def get_last(self) -> Optional[State]:
        """Get state that was added in the previous resolution round."""
        to_return = self._heap.get_last()  # type: Optional[State]
        return to_return

    def get_random(self) -> State:
        """Get a random state from beam."""
        idx = random.randint(0, self.size - 1) if self.size > 1 else 0
        return self.get(idx)

    def remove(self, state: State) -> None:
        """Remove the given state from beam."""
        try:
            self._heap.remove(state)
        except ValueError:  # TODO: fix
            _LOGGER.exception(
                "Encountered exception reported in https://github.com/thoth-station/adviser/issues/1541, ignoring..."
            )

    def pop(self, idx: Optional[int] = None) -> State:
        """Pop i-th element from the beam and remove it from the beam (this is actually toppop).

        If index is not provided, pop the largest item kept in the beam.
        """
        if idx is None:
            to_pop_state = self._heap.get_max()  # type: State
        else:
            to_pop_state = self._heap.get(idx)

        self._heap.remove(to_pop_state)
        return to_pop_state



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation of Temporal Difference (TD) based predictor with adaptive simulated annealing schedule."""

from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Optional
import logging
import math
import random
import signal

import attr

from .annealing import AdaptiveSimulatedAnnealing
from ..state import State


_LOGGER = logging.getLogger(__name__)
_INSTANCE = None


def _sigusr1_handler(sig_num: int, _: Any) -> None:
    """Handle SIGUSR1 that switches the predictor to exploitation (soon resolution timeout)."""
    global _INSTANCE

    _LOGGER.debug("Switching to exploitation phase based on a signal")
    _INSTANCE._temperature = 0.0  # type: ignore


@attr.s(slots=True)
class TemporalDifference(AdaptiveSimulatedAnnealing):
    """Implementation of Temporal Difference (TD) based predictor with adaptive simulated annealing schedule."""

    step = attr.ib(type=int, default=1, kw_only=True)
    trace = attr.ib(type=bool, default=True, kw_only=True)
    _policy = attr.ib(type=Dict[Tuple[str, str, str], List[Union[float, int]]], factory=dict, init=False)
    _steps_reward = attr.ib(type=float, default=0.0, init=False)
    _steps_taken = attr.ib(type=int, default=0, init=False)
    _next_state = attr.ib(type=Optional[State], default=None, init=False)
    _old_handler = attr.ib(type=Any, init=False, default=None)

    @step.validator
    def _step_validator(self, _: str, value: int) -> None:
        """Validate step parameter for n-step TD-learning."""
        if not isinstance(value, int):
            raise ValueError(f"Unknown type for TD n-step: {type(value)}")

        if value < 1:
            raise ValueError(f"Step set to {value} is not valid for n-step TD-learning")

    def pre_run(self) -> None:
        """Initialize pre-running of this predictor."""
        global _INSTANCE

        super().pre_run()
        self._policy.clear()
        self._temperature = float(self.context.limit)
        self._steps_taken = 0
        self._steps_reward = 0.0
        self._next_state = None

        # Setup handler that responds to SIGUSR1 and switches to exploitation phase.
        self._old_handler = signal.getsignal(signal.SIGUSR1)
        signal.signal(signal.SIGUSR1, _sigusr1_handler)
        _INSTANCE = self

    def post_run(self) -> None:
        """De-initialize resources used by this predictor."""
        global _INSTANCE

        signal.signal(signal.SIGUSR1, self._old_handler)
        self._old_handler = None
        _INSTANCE = None

    def set_reward_signal(self, state: State, package_tuple: Tuple[str, str, str], reward: float) -> None:
        """Note down reward signal of the last action performed."""
        trajectory_end = math.isnan(reward) or math.isinf(reward)
        if trajectory_end:
            self._next_state = None
            reward = 0.0

        self._steps_reward += reward
        if self._steps_taken < self.step and not trajectory_end:
            return

        if self.trace:
            for package_tuple in state.iter_resolved_dependencies():
                record = self._policy.setdefault(package_tuple, [0.0, 0])
                record[0] += self._steps_reward
                record[1] += 1
        else:
            record = self._policy.setdefault(package_tuple, [0.0, 0])
            record[0] += self._steps_reward
            record[1] += 1

        self._steps_taken = 0  # Set back to zero as we update policy.
        self._steps_reward = 0.0

    def run(self) -> Tuple[State, Tuple[str, str, str]]:
        """Run Temporal Difference (TD) with adaptive simulated annealing schedule."""
        if self._next_state is not None:
            unresolved_dependency_tuple = self._next_state.get_random_unresolved_dependency(prefer_recent=True)
            if self.keep_history:
                self._temperature_history.append((None, None, None, self.context.accepted_final_states_count))
            return self._next_state, unresolved_dependency_tuple

        if self._temperature > 0.0:
            self._temperature = self._temperature_function(self._temperature, self.context)
            self._next_state = self.context.beam.max()

            # Pick a random state to be expanded if accepted.
            probable_state_idx = random.randrange(1, self.context.beam.size) if self.context.beam.size > 1 else 0
            probable_state = self.context.beam.get(probable_state_idx)
            acceptance_probability = self._compute_acceptance_probability(
                self._next_state.score, probable_state.score, self._temperature
            )

            if acceptance_probability >= random.random():
                # Perform exploration.
                unresolved_dependency_tuple = probable_state.get_random_unresolved_dependency(prefer_recent=True)
                self._next_state = probable_state
            else:
                unresolved_dependency_tuple = self._do_exploitation(self._next_state)

            if self.keep_history:
                self._temperature_history.append(
                    (
                        self._temperature,
                        self._next_state is self.context.beam.max(),
                        acceptance_probability,
                        self.context.accepted_final_states_count,
                    )
                )

            self._steps_taken += 1
            return self._next_state, unresolved_dependency_tuple

        self._steps_taken += 1
        self._next_state = self.context.beam.max()
        if self.keep_history:
            self._temperature_history.append(
                (
                    self._temperature,
                    True,
                    0.0,
                    self.context.accepted_final_states_count,
                )
            )

        return self._next_state, self._do_exploitation(self._next_state)

    def _do_exploitation(self, state: State) -> Tuple[str, str, str]:
        """Perform expansion of a highest rated stack with action that should yield highest reward."""
        to_resolve_average = None
        to_resolve_package_tuple = None
        for package_tuple in state.iter_unresolved_dependencies():
            reward_records = self._policy.get(package_tuple)
            if reward_records is None:
                continue

            # Compute average - we want to be skewed based on the reward signal
            # we aggregate (so for example median of medians is not that suitable).
            average = reward_records[0] / reward_records[1]
            if to_resolve_average is None or to_resolve_average < average:
                to_resolve_average = average
                to_resolve_package_tuple = package_tuple

        # Make sure we found a candidate based on rewards marked. If not, pick a random one.
        return to_resolve_package_tuple or state.get_random_unresolved_dependency(prefer_recent=True)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve that makes sure packages shipped inside an s2i container are reused.

An example can be a Jupyter Notebook image that has TensorFlow pre-installed. As
this image is supposed to be "supported", TensorFlow that is present should be
reused, rather than re-installed to another version.
"""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion
from voluptuous import Schema
from voluptuous import Required

from ..sieve import Sieve

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class ThothS2IPackagesSieve(Sieve):
    """Remove packages that are already present in s2i image."""

    CONFIGURATION_DEFAULT = {"package_name": None}
    CONFIGURATION_SCHEMA: Schema = Schema({Required("package_name"): str, Required("package_version"): str})

    _JUSTIFICATION = jl("s2i_packages")
    _THOTH_S2I_IMAGE_PREFIX = "quay.io/thoth-station/"
    _THOTH_S2I_PACKAGES_LOCATION_PREFIX = "/opt/app-root/"

    _message_logged = attr.ib(type=bool, init=False)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register if the base image provided is Thoth's s2i."""
        if builder_context.is_included(cls):
            yield from ()
            return None

        base_image = builder_context.project.runtime_environment.base_image
        if not base_image or not base_image.startswith(cls._THOTH_S2I_IMAGE_PREFIX):
            yield from ()
            return None

        base_image_parts = cls.get_base_image(base_image)
        if not base_image_parts:
            _LOGGER.debug(
                "Failed to parse base image parts from %r, any possible Python packages available in "
                "the image will not be considered",
                base_image,
            )
            yield from ()
            return None

        analysis_document_id = builder_context.graph.get_last_analysis_document_id(
            base_image_parts[0],
            base_image_parts[1],
            is_external=False,
        )

        python_packages = builder_context.graph.get_python_package_version_all(analysis_document_id)
        if not python_packages:
            _LOGGER.debug(
                "No Python packages found in %r, not considering any possible Python packages available in %r",
                analysis_document_id,
                base_image,
            )
            yield from ()
            return None

        for python_package in python_packages:
            if not python_package["location"].startswith(cls._THOTH_S2I_PACKAGES_LOCATION_PREFIX):
                continue

            _LOGGER.debug(
                "Taking into account package %r in version %r located in %r in the container image",
                python_package["package_name"],
                python_package["package_version"],
                python_package["location"],
            )
            yield {"package_name": python_package["package_name"], "package_version": python_package["package_version"]}

        return None

    def pre_run(self) -> None:
        """Initialize before running the pipeline unit."""
        self._message_logged = False
        super().pre_run()

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """If a package does not meet version already present in the base image, remove it."""
        for package_version in package_versions:
            if package_version.locked_version == self.configuration["package_version"]:
                yield package_version
                continue

            if not self._message_logged:
                self._message_logged = True

                message = (
                    f"Removing any possible versions of {self.configuration['package_name']} as "
                    f"the given package is already present in the base image "
                    f"in version {self.configuration['package_version']}"
                )
                _LOGGER.warning("%s - see %s", message, self._JUSTIFICATION)
                self.context.stack_info.append({"type": "WARNING", "message": message, "link": self._JUSTIFICATION})



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test sieve for computing package combinations."""

import pytest

from thoth.adviser.context import Context
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import PackageCombinationsSieve
from thoth.python import Source
from thoth.python import PackageVersion

from ..base import AdviserUnitTestCase


class TestPackageCombinationsSieve(AdviserUnitTestCase):
    """Test sieve for computing package combinations."""

    UNIT_TESTED = PackageCombinationsSieve

    @pytest.mark.skip(reason="Package combinations sieve is never registered.")
    def test_verify_multiple_should_include(self) -> None:
        """Verify multiple should_include calls do not loop endlessly."""

    def test_no_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test obtaining default configuration."""
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    def test_pre_run(self, context: Context) -> None:
        """Test initializing before each run."""
        unit = self.UNIT_TESTED()
        unit.update_configuration({"package_name": None, "package_combinations": ["flask", "werkzeug"]})
        unit._package_tuples_seen = {"flask": ("flask", "0.12", "https://pypi.org/simple")}
        unit._package_combinations = {"foo", "bar"}

        unit.pre_run()

        assert unit._package_combinations == {"flask", "werkzeug"}
        assert unit._package_tuples_seen == {}

    @pytest.mark.skip(reason="Call to pre_run tested below.")
    def test_super_pre_run(self, context: Context) -> None:
        """This test case is tested below."""

    def test_run(self) -> None:
        """Test computing package combinations."""
        pypi = Source("https://pypi.org/simple")
        pv0 = PackageVersion(
            name="flask",
            version="==1.0.0",
            index=pypi,
            develop=False,
        )
        pv1 = PackageVersion(
            name="flask",
            version="==0.12",
            index=pypi,
            develop=False,
        )
        pv2 = PackageVersion(
            name="pandas",
            version="==1.3.3",
            index=pypi,
            develop=False,
        )
        pv3 = PackageVersion(
            name="pandas",
            version="==1.0.0",
            index=pypi,
            develop=False,
        )

        unit = self.UNIT_TESTED()
        unit.update_configuration(
            {
                "package_name": None,
                "package_combinations": ["flask", "werkzeug"],
            }
        )
        unit.unit_run = True
        unit.pre_run()

        assert unit.unit_run is False, "Parent pre_run was probably not called"

        assert list(unit.run((pv for pv in (pv0,)))) == [pv0]
        assert list(unit.run((pv for pv in (pv1,)))) == [pv1]
        assert list(unit.run((pv for pv in (pv2,)))) == [pv2]
        assert list(unit.run((pv for pv in (pv3,)))) == []
        assert list(unit.run((pv for pv in (pv2,)))) == [pv2]
        assert list(unit.run((pv for pv in (pv1,)))) == [pv1]
        assert list(unit.run((pv for pv in (pv0,)))) == [pv0]

        assert unit._package_combinations == {"flask", "werkzeug"}
        assert unit._package_tuples_seen == {
            "pandas": pv2.to_tuple(),
        }



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve to filter out packages respecting desired combinations that should be computed."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Set
from typing import Tuple
from typing import Union
from typing import TYPE_CHECKING

import attr
from thoth.python import PackageVersion
from voluptuous import Required
from voluptuous import Schema

from ..sieve import Sieve

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class PackageCombinationsSieve(Sieve):
    """A sieve to filter out packages respecting desired combinations that should be computed."""

    CONFIGURATION_DEFAULT: Dict[str, Union[None, List[str]]] = {"package_name": None, "package_combinations": []}
    CONFIGURATION_SCHEMA = Schema(
        {
            Required("package_name"): None,
            Required("package_combinations"): [str],
        }
    )

    _package_tuples_seen = attr.ib(
        type=Dict[str, Tuple[str, str, str]], default=attr.Factory(dict), kw_only=True, init=False
    )
    _package_combinations = attr.ib(type=Set[str], default=attr.Factory(set), kw_only=True, init=False)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self, never."""
        yield from ()
        return None

    def pre_run(self) -> None:
        """Prepare for the actual run."""
        self._package_combinations = set(self.configuration["package_combinations"])
        if not self._package_combinations:
            raise NotImplementedError("No package packages to compute combinations supplied")

        self._package_tuples_seen.clear()
        super().pre_run()

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Filter out packages based on Python package index configured for the given package."""
        for package_version in package_versions:
            if package_version.name in self._package_combinations:
                # Yield any version of this package based on the configuration.
                yield package_version
                continue

            seen_package_tuple = self._package_tuples_seen.get(package_version.name)
            if seen_package_tuple is None:
                self._package_tuples_seen[package_version.name] = package_version.to_tuple()
                yield package_version
                continue

            if seen_package_tuple == package_version.to_tuple():
                # This version was already used in one of the previous runs.
                yield package_version
                continue



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Kevin Postlethwait
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test Security Indicator step."""

from flexmock import flexmock
import pytest

from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.steps import SecurityIndicatorStep
from thoth.adviser.exceptions import NotAcceptable
from thoth.python import PackageVersion
from thoth.python import Source
from thoth.storages import GraphDatabase
from thoth.storages.exceptions import NotFoundError

from ..base import AdviserUnitTestCase


class TestSecurityIndicatorStep(AdviserUnitTestCase):
    """Test different aspects of si pipeline step."""

    UNIT_TESTED = SecurityIndicatorStep

    _SECURITY_INFO_EXISTS = {
        "severity_high_confidence_high": 0,
        "severity_high_confidence_medium": 5,
        "severity_high_confidence_low": 3,
        "severity_medium_confidence_high": 2,
        "severity_medium_confidence_medium": 0,
        "severity_medium_confidence_low": 8,
        "severity_low_confidence_high": 21,
        "severity_low_confidence_medium": 6,
        "severity_low_confidence_low": 0,
        "number_of_lines_with_code_in_python_files": 692,
    }

    _HIGH_HIGH_SECURITY_INFO = {
        "severity_high_confidence_high": 1,
        "severity_high_confidence_medium": 5,
        "severity_high_confidence_low": 3,
        "severity_medium_confidence_high": 2,
        "severity_medium_confidence_medium": 0,
        "severity_medium_confidence_low": 8,
        "severity_low_confidence_high": 21,
        "severity_low_confidence_medium": 6,
        "severity_low_confidence_low": 0,
        "number_of_lines_with_code_in_python_files": 692,
    }

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.recommendation_type = RecommendationType.SECURITY
        self.verify_multiple_should_include(builder_context)

    @pytest.mark.parametrize("recommendation_type", [RecommendationType.STABLE, RecommendationType.SECURITY])
    def test_include(self, builder_context: PipelineBuilderContext, recommendation_type: RecommendationType) -> None:
        """Test including this pipeline unit."""
        builder_context.decision_type = None
        builder_context.recommendation_type = recommendation_type
        assert builder_context.is_adviser_pipeline()
        assert list(SecurityIndicatorStep.should_include(builder_context)) == [{}]

    @pytest.mark.parametrize(
        "recommendation_type", [RecommendationType.LATEST, RecommendationType.PERFORMANCE, RecommendationType.TESTING]
    )
    def test_no_include(
        self,
        builder_context: PipelineBuilderContext,
        recommendation_type,
    ) -> None:
        """Test not including this pipeline unit step."""
        builder_context.decision_type = None
        builder_context.recommendation_type = recommendation_type
        assert list(SecurityIndicatorStep.should_include(builder_context)) == []

    def test_security_indicator_scoring(self) -> None:
        """Make sure we do score security indicators when the info is available."""
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_si_aggregated_python_package_version").with_args(
            package_name="flask", package_version="0.12.0", index_url="https://pypi.org/simple"
        ).and_return(self._SECURITY_INFO_EXISTS).once()

        package_version = PackageVersion(
            name="flask",
            version="==0.12.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context = flexmock(graph=GraphDatabase())
        context.recommendation_type = RecommendationType.STABLE
        with SecurityIndicatorStep.assigned_context(context):
            step = SecurityIndicatorStep()
            result = step.run(None, package_version)

        assert result is not None
        assert isinstance(result, tuple) and len(result) == 2
        assert isinstance(result[0], float)
        assert self.verify_justification_schema(result[1])

    @pytest.mark.parametrize("recommendation_type", [RecommendationType.SECURITY])
    def test_security_indicator_scoring_missing_secure(self, recommendation_type) -> None:
        """Make sure we don't accept package if si info is missing when recommendation is secure."""
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_si_aggregated_python_package_version").with_args(
            package_name="flask", package_version="0.12.0", index_url="https://pypi.org/simple"
        ).and_raise(NotFoundError).once()

        package_version = PackageVersion(
            name="flask",
            version="==0.12.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context = flexmock(graph=GraphDatabase(), stack_info=[])
        context.recommendation_type = recommendation_type
        with pytest.raises(NotAcceptable):
            with SecurityIndicatorStep.assigned_context(context):
                step = SecurityIndicatorStep()
                step.run(None, package_version)

        assert len(context.stack_info) == 1
        assert self.verify_justification_schema(context.stack_info)

    @pytest.mark.parametrize("recommendation_type", [RecommendationType.STABLE])
    def test_security_indicator_scoring_missing_stable(self, recommendation_type) -> None:
        """Make sure package is kept even if no score exists for security indicators and add justification."""
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_si_aggregated_python_package_version").with_args(
            package_name="flask", package_version="0.12.0", index_url="https://pypi.org/simple"
        ).and_raise(NotFoundError).once()

        package_version = PackageVersion(
            name="flask",
            version="==0.12.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context = flexmock(graph=GraphDatabase())
        context.recommendation_type = recommendation_type
        with SecurityIndicatorStep.assigned_context(context):
            step = SecurityIndicatorStep()
            result = step.run(None, package_version)

        assert result is not None
        assert isinstance(result, tuple) and len(result) == 2
        assert result[0] == 0
        assert len(result[1]) == 1
        assert self.verify_justification_schema(result[1])
        assert result[1][0]["type"] == "WARNING"
        assert (
            result[1][0]["message"] == "flask===0.12.0 on https://pypi.org/simple has no "
            "gathered information regarding security."
        )

    def test_security_indicator_with_high_confidence(self) -> None:
        """Make sure we don't accept package if si info is missing when recommendation is secure."""
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_si_aggregated_python_package_version").with_args(
            package_name="flask", package_version="0.12.0", index_url="https://pypi.org/simple"
        ).and_return(self._HIGH_HIGH_SECURITY_INFO).once()

        package_version = PackageVersion(
            name="flask",
            version="==0.12.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context = flexmock(graph=GraphDatabase(), stack_info=[])
        context.recommendation_type = RecommendationType.SECURITY
        with pytest.raises(NotAcceptable):
            with SecurityIndicatorStep.assigned_context(context):
                step = SecurityIndicatorStep()
                step.run(None, package_version)
        assert len(context.stack_info) == 1
        assert self.verify_justification_schema(context.stack_info)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of group step."""

import yaml

from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.prescription.v1 import GroupStepPrescription
from thoth.adviser.prescription.v1.schema import PRESCRIPTION_GROUP_STEP_SCHEMA

from .base import AdviserUnitPrescriptionTestCase


class TestGroupStepPrescription(AdviserUnitPrescriptionTestCase):
    """Test implementation of group step.

    As the run implementation is shared with raw step, only computation done in should_include is tested here.
    """

    def test_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test calculating configuration for the given prescription unit."""
        prescription_str = """\
name: GroupStep
type: step.Group
should_include:
  adviser_pipeline: true
match:
  group:
  - package_version:
      name: numpy
      version: '!=1.21.4'
      index_url:
        not: 'https://pypi.org/simple'
  - package_version:
      name: tensorflow
      version: '~=2.7.0'
  - package_version:
      name: pandas
      index_url: 'https://thoth-station.ninja'
      develop: true
  - package_version:
      name: werkzeug
      version: '==1.0.0'
      index_url: 'https://pypi.org/simple'
      develop: false
run:
  stack_info:
  - type: INFO
    message: Running the unit.
    link: 'https://thoth-station.ninja'
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_GROUP_STEP_SCHEMA(prescription)
        GroupStepPrescription.set_prescription(prescription)

        configurations = list(GroupStepPrescription.should_include(builder_context))

        assert configurations == [
            {
                "match": {
                    "package_version": {
                        "index_url": {"not": "https://pypi.org/simple"},
                        "name": "numpy",
                        "version": "!=1.21.4",
                    },
                    "state": {
                        "resolved_dependencies": [
                            {"name": "tensorflow", "version": "~=2.7.0"},
                            {"develop": True, "index_url": "https://thoth-station.ninja", "name": "pandas"},
                            {
                                "develop": False,
                                "index_url": "https://pypi.org/simple",
                                "name": "werkzeug",
                                "version": "==1.0.0",
                            },
                        ]
                    },
                },
                "multi_package_resolution": True,
                "package_name": "numpy",
                "prescription": {"run": False},
                "run": {
                    "stack_info": [
                        {"link": "https://thoth-station.ninja", "message": "Running the unit.", "type": "INFO"}
                    ]
                },
            },
            {
                "match": {
                    "package_version": {"name": "tensorflow", "version": "~=2.7.0"},
                    "state": {
                        "resolved_dependencies": [
                            {"index_url": {"not": "https://pypi.org/simple"}, "name": "numpy", "version": "!=1.21.4"},
                            {"develop": True, "index_url": "https://thoth-station.ninja", "name": "pandas"},
                            {
                                "develop": False,
                                "index_url": "https://pypi.org/simple",
                                "name": "werkzeug",
                                "version": "==1.0.0",
                            },
                        ]
                    },
                },
                "multi_package_resolution": True,
                "package_name": "tensorflow",
                "prescription": {"run": False},
                "run": {
                    "stack_info": [
                        {"link": "https://thoth-station.ninja", "message": "Running the unit.", "type": "INFO"}
                    ]
                },
            },
            {
                "match": {
                    "package_version": {"develop": True, "index_url": "https://thoth-station.ninja", "name": "pandas"},
                    "state": {
                        "resolved_dependencies": [
                            {"index_url": {"not": "https://pypi.org/simple"}, "name": "numpy", "version": "!=1.21.4"},
                            {"name": "tensorflow", "version": "~=2.7.0"},
                            {
                                "develop": False,
                                "index_url": "https://pypi.org/simple",
                                "name": "werkzeug",
                                "version": "==1.0.0",
                            },
                        ]
                    },
                },
                "multi_package_resolution": True,
                "package_name": "pandas",
                "prescription": {"run": False},
                "run": {
                    "stack_info": [
                        {"link": "https://thoth-station.ninja", "message": "Running the unit.", "type": "INFO"}
                    ]
                },
            },
            {
                "match": {
                    "package_version": {
                        "develop": False,
                        "index_url": "https://pypi.org/simple",
                        "name": "werkzeug",
                        "version": "==1.0.0",
                    },
                    "state": {
                        "resolved_dependencies": [
                            {"index_url": {"not": "https://pypi.org/simple"}, "name": "numpy", "version": "!=1.21.4"},
                            {"name": "tensorflow", "version": "~=2.7.0"},
                            {"develop": True, "index_url": "https://thoth-station.ninja", "name": "pandas"},
                        ]
                    },
                },
                "multi_package_resolution": True,
                "package_name": "werkzeug",
                "prescription": {"run": False},
                "run": {
                    "stack_info": [
                        {"link": "https://thoth-station.ninja", "message": "Running the unit.", "type": "INFO"}
                    ]
                },
            },
        ]

        for conf in configurations:
            GroupStepPrescription.CONFIGURATION_SCHEMA(conf)

    def test_should_include_groups(self, builder_context: PipelineBuilderContext) -> None:
        """Test calculating configuration for the given prescription unit, multiple groups."""
        prescription_str = """\
name: GroupStep
type: step.Group
should_include:
  adviser_pipeline: true
match:
- group:
  - package_version:
      name: numpy
      version: '!=1.21.4'
      index_url:
        not: 'https://pypi.org/simple'
  - package_version:
      name: tensorflow
      version: '~=2.7.0'
- group:
  - package_version:
      name: pandas
      index_url: 'https://thoth-station.ninja'
      develop: true
  - package_version:
      name: werkzeug
      version: '==1.0.0'
      index_url: 'https://pypi.org/simple'
      develop: false
run:
  stack_info:
  - type: INFO
    message: Running the unit.
    link: 'https://thoth-station.ninja'
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_GROUP_STEP_SCHEMA(prescription)
        GroupStepPrescription.set_prescription(prescription)

        configurations = list(GroupStepPrescription.should_include(builder_context))

        assert configurations == [
            {
                "match": {
                    "package_version": {
                        "index_url": {"not": "https://pypi.org/simple"},
                        "name": "numpy",
                        "version": "!=1.21.4",
                    },
                    "state": {"resolved_dependencies": [{"name": "tensorflow", "version": "~=2.7.0"}]},
                },
                "multi_package_resolution": True,
                "package_name": "numpy",
                "prescription": {"run": False},
                "run": {
                    "stack_info": [
                        {"link": "https://thoth-station.ninja", "message": "Running the unit.", "type": "INFO"}
                    ]
                },
            },
            {
                "match": {
                    "package_version": {"name": "tensorflow", "version": "~=2.7.0"},
                    "state": {
                        "resolved_dependencies": [
                            {"index_url": {"not": "https://pypi.org/simple"}, "name": "numpy", "version": "!=1.21.4"}
                        ]
                    },
                },
                "multi_package_resolution": True,
                "package_name": "tensorflow",
                "prescription": {"run": False},
                "run": {
                    "stack_info": [
                        {"link": "https://thoth-station.ninja", "message": "Running the unit.", "type": "INFO"}
                    ]
                },
            },
            {
                "match": {
                    "package_version": {"develop": True, "index_url": "https://thoth-station.ninja", "name": "pandas"},
                    "state": {
                        "resolved_dependencies": [
                            {
                                "develop": False,
                                "index_url": "https://pypi.org/simple",
                                "name": "werkzeug",
                                "version": "==1.0.0",
                            }
                        ]
                    },
                },
                "multi_package_resolution": True,
                "package_name": "pandas",
                "prescription": {"run": False},
                "run": {
                    "stack_info": [
                        {"link": "https://thoth-station.ninja", "message": "Running the unit.", "type": "INFO"}
                    ]
                },
            },
            {
                "match": {
                    "package_version": {
                        "develop": False,
                        "index_url": "https://pypi.org/simple",
                        "name": "werkzeug",
                        "version": "==1.0.0",
                    },
                    "state": {
                        "resolved_dependencies": [
                            {"develop": True, "index_url": "https://thoth-station.ninja", "name": "pandas"}
                        ]
                    },
                },
                "multi_package_resolution": True,
                "package_name": "werkzeug",
                "prescription": {"run": False},
                "run": {
                    "stack_info": [
                        {"link": "https://thoth-station.ninja", "message": "Running the unit.", "type": "INFO"}
                    ]
                },
            },
        ]

        for conf in configurations:
            GroupStepPrescription.CONFIGURATION_SCHEMA(conf)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019-2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation of sieves used in adviser pipeline."""

from .constraints import ConstraintsSieve
from .cve import CveSieve
from .experimental_filter_conf_index import FilterConfiguredIndexSieve
from .experimental_package_index import PackageIndexConfigurationSieve
from .experimental_prereleases import SelectiveCutPreReleasesSieve
from .filter_index import FilterIndexSieve
from .index_enabled import PackageIndexSieve
from .legacy_version import LegacyVersionSieve
from .locked import CutLockedSieve
from .package_combinations import PackageCombinationsSieve
from .prereleases import CutPreReleasesSieve
from ._pulp import NoPulpIndexLabelSieve
from ._pulp import PulpIndexLabelSieve
from .rules import SolverRulesSieve
from .solved import SolvedSieve
from .tensorflow import TensorFlowAPISieve
from .thoth_s2i_abi_compat import ThothS2IAbiCompatibilitySieve
from .thoth_s2i_packages import ThothS2IPackagesSieve
from .update import PackageUpdateSieve
from .version_constraint import VersionConstraintSieve

# Relative ordering of units is relevant, as the order specifies order
# in which the asked to be registered - any dependencies between them
# can be mentioned here.
__all__ = [
    "PackageUpdateSieve",
    "LegacyVersionSieve",
    "CutPreReleasesSieve",
    "PulpIndexLabelSieve",
    "NoPulpIndexLabelSieve",
    "ConstraintsSieve",
    "SelectiveCutPreReleasesSieve",
    "FilterConfiguredIndexSieve",
    "CveSieve",
    "PackageCombinationsSieve",
    "PackageIndexConfigurationSieve",
    "CutLockedSieve",
    "PackageIndexSieve",
    "SolvedSieve",
    "SolverRulesSieve",
    "VersionConstraintSieve",
    "ThothS2IPackagesSieve",
    "ThothS2IAbiCompatibilitySieve",
    "FilterIndexSieve",
    "TensorFlowAPISieve",
]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Kevin Postlethwait
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Score based on security indicators aggregated."""

from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING
import logging

import attr
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion
from thoth.storages.exceptions import NotFoundError
from voluptuous import Schema
from voluptuous import Required

from ..enums import RecommendationType
from ..exceptions import NotAcceptable
from ..state import State
from ..step import Step


if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class SecurityIndicatorStep(Step):
    """A step that scores a state based on security info aggregated."""

    _logged_packages = attr.ib(type=Set[Tuple[str, str, str]], default=attr.Factory(set), init=False)

    CONFIGURATION_DEFAULT = {
        "high_confidence_weight": 1.0,
        "high_severity_weight": 100.0,
        "low_confidence_weight": 0.1,
        "low_severity_weight": 1.0,
        "medium_confidence_weight": 0.5,
        "medium_severity_weight": 10.0,
        "multi_package_resolution": False,
        "package_name": None,
        "si_score_weight": 0.5,
        "function_scaling": 1 / 1000,
    }
    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("high_confidence_weight"): float,
            Required("high_severity_weight"): float,
            Required("low_confidence_weight"): float,
            Required("low_severity_weight"): float,
            Required("medium_confidence_weight"): float,
            Required("medium_severity_weight"): float,
            Required("multi_package_resolution"): bool,
            Required("package_name"): None,
            Required("si_score_weight"): float,
            Required("function_scaling"): float,
        }
    )
    _JUSTIFICATION_LINK_SECURITY = jl("security")
    _JUSTIFICATION_LINK_BANDIT = jl("si_bandit")

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register only if we are explicitly recommending secure stacks."""
        if builder_context.recommendation_type not in (RecommendationType.SECURITY, RecommendationType.STABLE):
            yield from ()
            return None

        if builder_context.is_included(cls):
            yield from ()
            return None

        yield {}
        return None

    @classmethod
    def _generate_justification(cls, name: str, version: str, index: str) -> List[Dict[str, Any]]:
        return [
            {
                "type": "WARNING",
                "message": f"{name}==={version} on {index} has no gathered information regarding security.",
                "link": cls._JUSTIFICATION_LINK_SECURITY,
                "package_name": name,
                "version_range": f"==={version}",
            }
        ]

    def pre_run(self) -> None:
        """Initialize this pipeline step before running the pipeline."""
        self._logged_packages.clear()
        super().pre_run()

    def run(
        self, state: State, package_version: PackageVersion
    ) -> Optional[Tuple[Optional[float], Optional[List[Dict[str, str]]]]]:
        """Score package based on security indicators gathered, do not include if not analyzed."""
        package_version_tuple = package_version.to_tuple_locked()
        justification = []
        try:
            s_info = self.context.graph.get_si_aggregated_python_package_version(
                package_name=package_version.name,
                package_version=package_version.locked_version,
                index_url=package_version.index.url,
            )
            msg = (
                f"Thoth has security info for {package_version.name}==={package_version.locked_version} "
                f"on {package_version.index.url}"
            )
            justification.append(
                {
                    "type": "INFO",
                    "message": msg,
                    "link": self._JUSTIFICATION_LINK_SECURITY,
                    "package_name": package_version.name,
                    "version_range": f"==={package_version.locked_version}",
                }
            )
        except NotFoundError:
            if self.context.recommendation_type == RecommendationType.SECURITY:
                if package_version_tuple not in self._logged_packages:
                    self._logged_packages.add(package_version_tuple)
                    msg = (
                        f"No security info for {package_version.name}==={package_version.locked_version} "
                        f"on {package_version.index.url}"
                    )
                    self.context.stack_info.append(
                        {
                            "type": "WARNING",
                            "message": msg,
                            "link": self._JUSTIFICATION_LINK_SECURITY,
                            "package_name": package_version.name,
                            "version_range": f"==={package_version.locked_version}",
                        }
                    )
                    _LOGGER.warning("%s - see %s", msg, self._JUSTIFICATION_LINK_SECURITY)
                raise NotAcceptable
            return (
                0,
                self._generate_justification(
                    name=package_version.name, version=package_version.locked_version, index=package_version.index.url
                ),
            )

        if (
            self.context.recommendation_type == RecommendationType.SECURITY
            and s_info["severity_high_confidence_high"] != 0
        ):
            if package_version_tuple not in self._logged_packages:
                self._logged_packages.add(package_version_tuple)
                msg = (
                    f"Skipping including package {package_version_tuple} because bandit found "
                    f"high severity high confidence issue(s)."
                )
                _LOGGER.warning("%s - %s", msg, self._JUSTIFICATION_LINK_BANDIT)
                self.context.stack_info.append(
                    {"type": "WARNING", "message": msg, "link": self._JUSTIFICATION_LINK_BANDIT}
                )
            raise NotAcceptable
        else:
            msg = f"bandit found no high severity, high confidence issues for {package_version_tuple}"
            justification.append({"type": "INFO", "message": msg, "link": self._JUSTIFICATION_LINK_BANDIT})

        s_score = (
            (
                s_info["severity_high_confidence_high"] * self.configuration["high_confidence_weight"]
                + s_info["severity_high_confidence_medium"] * self.configuration["medium_confidence_weight"]
                + s_info["severity_high_confidence_low"] * self.configuration["low_confidence_weight"]
            )
            * self.configuration["high_severity_weight"]
            + (
                s_info["severity_medium_confidence_high"] * self.configuration["high_confidence_weight"]
                + s_info["severity_medium_confidence_medium"] * self.configuration["medium_confidence_weight"]
                + s_info["severity_medium_confidence_low"] * self.configuration["low_confidence_weight"]
            )
            * self.configuration["medium_severity_weight"]
            + (
                s_info["severity_low_confidence_high"] * self.configuration["high_confidence_weight"]
                + s_info["severity_low_confidence_medium"] * self.configuration["medium_confidence_weight"]
                + s_info["severity_low_confidence_low"] * self.configuration["low_confidence_weight"]
            )
            * self.configuration["low_severity_weight"]
        )

        s_score = (
            (s_score / (s_info["number_of_lines_with_code_in_python_files"] or 1)) ** 2
            * self.configuration["function_scaling"]
            * -1
        )
        s_score = max(s_score, -1) * self.configuration["si_score_weight"]
        return s_score, justification or None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020-2021 Kevin Postlethwait
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test filtering out Python Packages based on required and available ABI symbols."""

from typing import Optional

from flexmock import flexmock
import pytest

from thoth.adviser.context import Context
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import ThothS2IAbiCompatibilitySieve
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion
from thoth.python import Source
from thoth.storages import GraphDatabase

from ..base import AdviserUnitTestCase

_SYSTEM_SYMBOLS = ["GLIBC_2.0", "GLIBC_2.1", "GLIBC_2.2", "GLIBC_2.3", "GLIBC_2.4", "GLIBC_2.5", "GCC_3.4", "X_2.21"]
_REQUIRED_SYMBOLS_A = ["GLIBC_2.9"]
_REQUIRED_SYMBOLS_B = ["GLIBC_2.4"]


class TestThothS2IAbiCompatibilitySieve(AdviserUnitTestCase):
    """Test filtering out packages based on symbols required."""

    UNIT_TESTED = ThothS2IAbiCompatibilitySieve

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38:v0.23.0"
        self.verify_multiple_should_include(builder_context)

    @pytest.mark.parametrize(
        "base_image",
        [
            None,
            "fedora:32",
        ],
    )
    def test_no_should_include(self, base_image: Optional[str], builder_context: PipelineBuilderContext) -> None:
        """Test not including this pipeline unit."""
        builder_context.project.runtime_environment.base_image = base_image
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    def test_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test including this pipeline unit."""
        builder_context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38:v0.23.0"
        assert list(self.UNIT_TESTED.should_include(builder_context)) == [{}]

    def test_no_thoth_s2i_version(self, context: Context) -> None:
        """Test no Thoth S2I version present."""
        context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38"  # No version.

        GraphDatabase.should_receive("get_thoth_s2i_analyzed_image_symbols_all").times(0)

        unit = self.UNIT_TESTED()
        with self.UNIT_TESTED.assigned_context(context):
            unit.pre_run()

        assert not unit.image_symbols

    def test_abi_compat_symbols_present(self, context: Context) -> None:
        """Test if required symbols are correctly identified."""
        source = Source("https://pypi.org/simple")
        package_version = PackageVersion(name="tensorflow", version="==1.9.0", index=source, develop=False)
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_thoth_s2i_analyzed_image_symbols_all").and_return(_SYSTEM_SYMBOLS).once()
        GraphDatabase.should_receive("get_python_package_required_symbols").and_return(_REQUIRED_SYMBOLS_B).once()

        context.project.runtime_environment.operating_system.name = "rhel"
        context.project.runtime_environment.operating_system.version = "8.0"
        context.project.runtime_environment.cuda_version = "4.6"
        context.project.runtime_environment.python_version = "3.6"
        context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38:v0.23.0"

        with self.UNIT_TESTED.assigned_context(context):
            sieve = self.UNIT_TESTED()
            sieve.pre_run()
            assert list(sieve.run((p for p in [package_version]))) == [package_version]

    def test_abi_compat_symbols_not_present(self, context: Context) -> None:
        """Test if required symbols being missing is correctly identified."""
        source = Source("https://pypi.org/simple")
        package_version = PackageVersion(name="tensorflow", version="==1.9.0", index=source, develop=False)
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_thoth_s2i_analyzed_image_symbols_all").and_return(_SYSTEM_SYMBOLS).once()
        GraphDatabase.should_receive("get_python_package_required_symbols").and_return(_REQUIRED_SYMBOLS_A).once()

        context.project.runtime_environment.operating_system.name = "rhel"
        context.project.runtime_environment.operating_system.version = "8.0"
        context.project.runtime_environment.cuda_version = "4.6"
        context.project.runtime_environment.python_version = "3.6"
        context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38:v0.23.0"

        with self.UNIT_TESTED.assigned_context(context):
            sieve = self.UNIT_TESTED()
            sieve.pre_run()
            assert list(sieve.run((p for p in [package_version]))) == []

    def test_super_pre_run(self, context: Context) -> None:
        """Make sure the pre-run method of the base is called."""
        context.graph.should_receive("get_thoth_s2i_analyzed_image_symbols_all").with_args(
            thoth_s2i_image_name="quay.io/thoth-station/s2i-thoth-ubi8-py38",
            thoth_s2i_image_version="0.23.0",
            is_external=False,
        ).and_return(set()).once()

        context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38:v0.23.0"

        unit = self.UNIT_TESTED()
        assert unit.unit_run is False

        with unit.assigned_context(context):
            unit.pre_run()

        assert unit.unit_run is False

    def test_pre_run_no_symbols_warning(self, context: Context) -> None:
        """Make sure the pre_run method produces a warning for users."""
        context.graph.should_receive("get_thoth_s2i_analyzed_image_symbols_all").with_args(
            thoth_s2i_image_name="quay.io/thoth-station/s2i-thoth-ubi8-py38",
            thoth_s2i_image_version="1.0.0",
            is_external=False,
        ).and_return(set()).once()

        context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38:v1.0.0"

        assert not context.stack_info

        unit = self.UNIT_TESTED()
        with unit.assigned_context(context):
            unit.pre_run()

        assert len(context.stack_info) == 1
        assert self.verify_justification_schema(context.stack_info)
        assert context.stack_info == [
            {
                "link": jl("no_abi"),
                "message": "No ABI symbols found for " "'quay.io/thoth-station/s2i-thoth-ubi8-py38' in version '1.0.0'",
                "type": "WARNING",
            },
        ]

    def test_run_no_symbols(self, context: Context) -> None:
        """Make sure that no information about symbols available in the image do not adjust the resolution process."""
        context.graph.should_receive("get_thoth_s2i_analyzed_image_symbols_all").with_args(
            thoth_s2i_image_name="quay.io/thoth-station/s2i-thoth-ubi8-py38",
            thoth_s2i_image_version="1.0.0",
            is_external=False,
        ).and_return(set()).once()

        pv = PackageVersion(
            name="flask",
            version="==1.0.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38:v1.0.0"

        unit = self.UNIT_TESTED()
        with unit.assigned_context(context):
            unit.pre_run()
            assert list(unit.run((pv for pv in (pv,)))) == [pv]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""File conftest.py for pytest test suite."""

from typing import Callable

import pytest
from flexmock import flexmock

from thoth.adviser.context import Context
from thoth.adviser.beam import Beam
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.pipeline_config import PipelineConfig
from thoth.adviser.resolver import Resolver
from thoth.adviser.predictor import Predictor
from thoth.adviser.prescription import UnitPrescription
from thoth.adviser.state import State
from thoth.common import RuntimeEnvironment
from thoth.python import Project

from .units.boots import Boot1
from .units.pseudonyms import Pseudonym1
from .units.sieves import Sieve1
from .units.steps import Step1
from .units.strides import Stride1
from .units.wraps import Wrap1

from thoth.storages import GraphDatabase

from .base import AdviserTestCase


class PredictorMock(Predictor):
    """A mocked predictor for testing, which always returns a random index to beam."""

    def run(self) -> int:
        """Run the main method used for predictor."""
        return self.context.beam.get_random()


@pytest.fixture
def context(project: Project) -> Context:
    """Create a fixture for a clean context."""
    flexmock(Context)
    flexmock(Beam)
    flexmock(GraphDatabase)

    return Context(
        project=project,
        graph=GraphDatabase(),
        library_usage=None,
        labels={},
        limit=100,
        count=3,
        beam=Beam(),
        recommendation_type=RecommendationType.LATEST,
    )


@pytest.fixture
def pipeline_config() -> PipelineConfig:
    """Create a fixture for a pipeline configuration with few representatives of each pipeline unit type."""
    flexmock(PipelineConfig)

    flexmock(Boot1)
    flexmock(Pseudonym1)
    flexmock(Sieve1)
    flexmock(Step1)
    flexmock(Stride1)
    flexmock(Wrap1)

    boot1 = Boot1()
    pseudonym1 = Pseudonym1()
    sieve1 = Sieve1()
    step1 = Step1()
    stride1 = Stride1()
    wrap1 = Wrap1()

    return PipelineConfig(
        boots={boot1.configuration.get("package_name"): [boot1]},
        pseudonyms={pseudonym1.configuration["package_name"]: [pseudonym1]},
        sieves={sieve1.configuration.get("package_name"): [sieve1]},
        steps={step1.configuration.get("package_name"): [step1]},
        strides={stride1.configuration.get("package_name"): [stride1]},
        wraps={wrap1.configuration.get("package_name"): [wrap1]},
    )


@pytest.fixture
def project() -> Project:
    """Create a fixture for a project representation."""
    flexmock(Project)
    flexmock(RuntimeEnvironment)

    pipfile_path = AdviserTestCase.data_dir / "projects" / "Pipfile"
    pipfile_lock_path = AdviserTestCase.data_dir / "projects" / "Pipfile.lock"

    return Project.from_files(
        pipfile_path=str(pipfile_path),
        pipfile_lock_path=str(pipfile_lock_path),
        runtime_environment=RuntimeEnvironment.from_dict({}),
    )


@pytest.fixture
def graph() -> GraphDatabase:
    """Create a knowledge graph connector fixture."""
    flexmock(GraphDatabase)

    graph = GraphDatabase()
    graph.connect()
    return graph


@pytest.fixture
def predictor_mock() -> Predictor:
    """Return a mock for predictor."""
    flexmock(PredictorMock)

    return PredictorMock()


@pytest.fixture
def predictor_mock_class() -> type:
    """Return a predictor mock class."""
    flexmock(PredictorMock)

    return PredictorMock


@pytest.fixture
def resolver(pipeline_config: PipelineConfig, project: Project, predictor_mock: Predictor) -> Resolver:
    """Create a resolver instance for tests."""
    flexmock(Resolver)
    flexmock(GraphDatabase)
    flexmock(Beam)

    return Resolver(
        pipeline=pipeline_config,
        project=project,
        library_usage={},
        graph=GraphDatabase(),
        predictor=predictor_mock,
        recommendation_type=RecommendationType.LATEST,
        limit=Resolver.DEFAULT_LIMIT,
        count=Resolver.DEFAULT_COUNT,
        beam_width=Resolver.DEFAULT_BEAM_WIDTH,
        limit_latest_versions=Resolver.DEFAULT_LIMIT_LATEST_VERSIONS,
    )


def _get_state() -> State:
    """Create a fixture for a non-final state."""
    unresolved_dependency = ("flask", "1.1.1", "https://pypi.org/simple")
    unresolved_dependencies = {id(unresolved_dependency): unresolved_dependency}
    state = State(
        score=0.1,
        unresolved_dependencies={"flask": unresolved_dependencies},
        resolved_dependencies={"hexsticker": ("hexsticker", "1.0.0", "https://pypi.org/simple")},
        advised_runtime_environment=RuntimeEnvironment.from_dict({"python_version": "3.6"}),
    )
    state.add_justification(AdviserTestCase.JUSTIFICATION_SAMPLE_1)
    return state


@pytest.fixture
def state() -> State:
    """Create a fixture for a non-final state."""
    return _get_state()


@pytest.fixture(scope="session")
def state_factory() -> Callable[[], State]:
    """Return a state - handy to be used in tests requiring hypothesis."""
    return _get_state


@pytest.fixture
def builder_context(project: Project) -> PipelineBuilderContext:
    """Create a fixture for a pipeline builder context."""
    flexmock(PipelineBuilderContext)
    flexmock(GraphDatabase)

    yield PipelineBuilderContext(
        graph=GraphDatabase(),
        project=project,
        library_usage=None,
        decision_type=None,
        recommendation_type=RecommendationType.LATEST,
    )
    # Clear any changes done to the should_include cache. This somehow mimics pipeline builder that
    # always clears the cache once the pipeline is constructed.
    UnitPrescription.SHOULD_INCLUDE_CACHE.clear()


@pytest.fixture(scope="session", autouse=True)
def state_verify_justification_schema() -> None:
    """Set automatic checks for schema validation for justifications added to states by pipeline units."""
    # Patch the justification function.
    def _patched_add_justification(self, justification):
        """Patch State.add_justification."""
        AdviserTestCase.verify_justification_schema(justification)
        add_justification_func(self, justification)

    add_justification_func = State.add_justification
    State.add_justification = _patched_add_justification



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Various helpers and utility functions."""

import os
import logging

from typing import Any
from typing import Set


def should_keep_history(value: Any) -> bool:
    """Check if the history should be kept.

    Used as click's converter. If not set explicitly during invocation, check
    environment variable to turn off history tracking. The default value of
    `value' is None which triggers checks in environment variables.
    """
    if value is None:
        return bool(int(os.getenv("THOTH_ADVISER_HISTORY", 0)))

    if isinstance(value, bool):
        return value

    raise ValueError(f"Unknown keep history configuration value: {value!r} if of type {type(value)!r}")


def log_once(
    logger: logging.Logger,
    log_state: Set[Any],
    log_state_key: Any,
    msg: str,
    *args: Any,
    level: int = logging.WARNING,
    **kwargs: Any,
) -> None:
    """Log the given message once."""
    if log_state_key in log_state:
        # Already logged, noop.
        return

    log_state.add(log_state_key)
    logger.log(level, msg, *args, **kwargs)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A base class for implementing package pseudonyms."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Tuple
from typing import TYPE_CHECKING

from thoth.python import PackageVersion
from voluptuous import Schema
from voluptuous import Required
import attr

from packaging.specifiers import SpecifierSet
from .unit import UnitPrescription
from .schema import PRESCRIPTION_PSEUDONYM_MATCH_ENTRY_SCHEMA
from .schema import PRESCRIPTION_PSEUDONYM_RUN_SCHEMA

if TYPE_CHECKING:
    from ...pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class PseudonymPrescription(UnitPrescription):
    """Pseudonym base class implementation."""

    # Pseudonym is always specific to a package.
    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("package_name"): str,
            Required("match"): PRESCRIPTION_PSEUDONYM_MATCH_ENTRY_SCHEMA,
            Required("run"): PRESCRIPTION_PSEUDONYM_RUN_SCHEMA,
            Required("prescription"): Schema({"run": bool}),
        }
    )

    _logged = attr.ib(type=bool, kw_only=True, init=False, default=False)
    _specifier = attr.ib(type=Optional[SpecifierSet], kw_only=True, init=False, default=None)
    _index_url = attr.ib(type=Optional[str], kw_only=True, init=False, default=None)

    @staticmethod
    def is_pseudonym_unit_type() -> bool:
        """Check if this unit is of type pseudonym."""
        return True

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if the given pipeline unit should be included in the given pipeline configuration."""
        if cls._should_include_base(builder_context):
            prescription: Dict[str, Any] = cls._PRESCRIPTION  # type: ignore
            yield from cls._yield_should_include(prescription)
            return None

        yield from ()
        return None

    def pre_run(self) -> None:
        """Prepare before running this pipeline unit."""
        package_version = self.match_prescription.get("package_version", {})
        version_specifier = package_version.get("version")
        if version_specifier:
            self._specifier = SpecifierSet(version_specifier)
            self._specifier.prereleases = True

        self._index_url = package_version.get("index_url")
        self._logged = False
        super().pre_run()

    def run(self, package_version: PackageVersion) -> Generator[Tuple[str, str, str], None, None]:
        """Run main entry-point for pseudonyms to map packages to their counterparts."""
        if (not self._index_url_check(self._index_url, package_version.index.url)) or (
            self._specifier is not None and package_version.locked_version not in self._specifier
        ):
            yield from ()
            return None

        to_yield = self.run_prescription["yield"]
        to_yield_package_version = to_yield.get("package_version") or {}
        if to_yield.get("yield_matched_version"):
            pseudonym_package_version = package_version.locked_version
        else:
            pseudonym_package_version = to_yield_package_version.get("locked_version")
            if pseudonym_package_version:
                pseudonym_package_version = pseudonym_package_version[2:]

        runtime_environment = self.context.project.runtime_environment
        pseudonyms = self.context.graph.get_solved_python_package_versions_all(
            package_name=to_yield_package_version.get("name"),
            package_version=pseudonym_package_version,
            index_url=to_yield_package_version.get("index_url"),
            count=None,
            os_name=runtime_environment.operating_system.name,
            os_version=runtime_environment.operating_system.version,
            python_version=runtime_environment.python_version,
            distinct=True,
            is_missing=False,
        )

        prescription_conf = self._configuration["prescription"]
        if pseudonyms and not prescription_conf["run"]:
            self._run_stack_info()
            self._run_log()
            prescription_conf["run"] = True

        for pseudonym in pseudonyms:
            _LOGGER.info(
                "%s: Considering package %r as a pseudonym of %r",
                self.get_unit_name(),
                pseudonym,
                package_version.to_tuple(),
            )
            yield pseudonym[0], pseudonym[1], pseudonym[2]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to provide a link to Thoth search showing results in a UI."""

import logging
import os
from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

import attr

from ..boot import Boot

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)

_THOTH_SEARCH_URL = os.getenv(
    "THOTH_SEARCH_ADVISER_URL", "https://thoth-station.ninja/search/advise/{document_id}/summary"
)
_DOCUMENT_ID = os.getenv("THOTH_DOCUMENT_ID", "UNKNOWN")


@attr.s(slots=True)
class ThothSearchBoot(Boot):
    """A boot to provide a link to Thoth search showing results in a UI."""

    _search_url = attr.ib(type=str, default=_THOTH_SEARCH_URL, kw_only=True)
    _document_id = attr.ib(type=str, default=_DOCUMENT_ID, kw_only=True)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self, once, always."""
        if builder_context.is_adviser_pipeline() and not builder_context.is_included(cls):
            yield {}
            return None

        yield from ()
        return None

    def run(self) -> None:
        """Provide a link to Thoth search visualizing results."""
        self.context.stack_info.append(
            {
                "type": "INFO",
                "message": "Results can be browsed in Thoth search",
                "link": self._search_url.format(document_id=self._document_id),
            }
        )



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of step prescription v1."""

from flexmock import flexmock
import pytest
import yaml

from thoth.adviser.context import Context
from thoth.adviser.state import State
from thoth.adviser.prescription.v1 import StepPrescription
from thoth.adviser.prescription.v1.schema import PRESCRIPTION_STEP_SCHEMA
from thoth.python import PackageVersion
from thoth.python import Source

from .base import AdviserUnitPrescriptionTestCase


class TestStepPrescription(AdviserUnitPrescriptionTestCase):
    """Tests related to step prescription v1."""

    def test_run_stack_info(self, context: Context, state: State) -> None:
        """Check assigning stack info."""
        prescription_str = """
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    index_url: 'https://thoth-station.ninja/simple'
run:
  stack_info:
    - type: WARNING
      message: Some message
      link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="flask",
            version="==1.1.0",
            index=Source("https://thoth-station.ninja/simple"),
            develop=False,
        )

        self.check_run_stack_info(context, StepPrescription, state=state, package_version=package_version)

    @pytest.mark.parametrize("log_level", ["INFO", "ERROR", "WARNING"])
    def test_run_log(self, caplog, context: Context, state: State, log_level: str) -> None:
        """Check logging messages."""
        prescription_str = f"""
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: flask
run:
  log:
    message: Seen flask during resolution
    type: {log_level}
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="flask",
            version="==1.1.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )
        self.check_run_log(caplog, context, log_level, StepPrescription, state=state, package_version=package_version)

    def test_run_eager_stop_pipeline(self, context: Context, state: State) -> None:
        """Check eager stop pipeline configuration."""
        prescription_str = """
name: StepUnit
type: step
should_include:
  times: 1
  dependency_monkey_pipeline: true
match:
  package_version:
    name: flask
    version: "<1.0.0"
run:
  eager_stop_pipeline: This is exception message reported
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="flask",
            version="==0.12",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )
        self.check_run_eager_stop_pipeline(context, StepPrescription, state=state, package_version=package_version)

    def test_run_not_acceptable(self, context: Context, state: State) -> None:
        """Check raising not acceptable."""
        prescription_str = """
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
  dependency_monkey_pipeline: true
match:
  package_version:
      name: flask
      version: "~=0.0"
      index_url: "https://pypi.org/simple"
run:
  not_acceptable: This is exception message reported
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="flask",
            version="==0.12",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )
        self.check_run_not_acceptable(context, StepPrescription, state=state, package_version=package_version)

    def test_run(self, context: Context, state: State) -> None:
        """Check running the step with score and justification."""
        prescription_str = """
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: pysaml2
    version: '<6.5.0'
    index_url: 'https://pypi.org/simple'
run:
  score: -0.1
  justification:
    - type: WARNING
      message: CVE found for pysaml2
      link: cve_pysaml2
    - type: INFO
      message: Package pysaml2 was removed from software stack resolution
      link: https://example.com
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="pysaml2",
            version="==6.4.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        unit = StepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = unit.run(state, package_version)
            assert isinstance(result, tuple)
            assert len(result) == 2
            assert result[0] == -0.1
            assert result[1] == unit.run_prescription["justification"]

            # Run one more time to make sure justification is added only once.
            result = unit.run(state, package_version)
            assert isinstance(result, tuple)
            assert len(result) == 2
            assert result[0] == -0.1
            assert result[1] is None

    def test_run_state(self, context: Context, state: State) -> None:
        """Test running the prescription if state matches."""
        prescription_str = """
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
    version: '==1.19.1'
    index_url: 'https://pypi.org/simple'
  state:
    resolved_dependencies:
      - name: tensorflow
        version: '~=2.4.0'
run:
  score: 0.5
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        state.add_resolved_dependency(("tensorflow", "2.4.0", "https://pypi.org/simple"))

        unit = StepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = unit.run(state, package_version)

        assert isinstance(result, tuple)
        assert result[0] == 0.5
        assert result[1] is None

    def test_run_state_no_match(self, context: Context, state: State) -> None:
        """Test running the prescription if state matches."""
        prescription_str = """
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
    version: '==1.19.1'
    index_url: 'https://pypi.org/simple'
  state:
    resolved_dependencies:
      - name: tensorflow
        version: '~=2.4.0'
run:
  score: 0.5
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        state.add_resolved_dependency(("tensorflow", "2.2.0", "https://pypi.org/simple"))

        unit = StepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = unit.run(state, package_version)

        assert result is None

    def test_should_include_package_name(self) -> None:
        """Test including this pipeline unit."""
        prescription_str = """
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  - package_version:
      name: numpy
      version: '==1.19.1'
      index_url: 'https://pypi.org/simple'
run:
  multi_package_resolution: true
  score: 0.1
"""
        flexmock(StepPrescription).should_receive("_should_include_base").replace_with(lambda _: True).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)

        builder_context = flexmock()
        result = list(StepPrescription.should_include(builder_context))
        assert result == [
            {
                "package_name": "numpy",
                "multi_package_resolution": True,
                "prescription": {
                    "run": False,
                },
                "match": {
                    "package_version": {
                        "name": "numpy",
                        "version": "==1.19.1",
                        "index_url": "https://pypi.org/simple",
                    },
                },
                "run": {
                    "score": 0.1,
                    "multi_package_resolution": True,
                },
            }
        ]

    def test_should_include_no_package_name(self) -> None:
        """Test including this pipeline unit."""
        prescription_str = """
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    index_url: 'https://thoth-station.ninja'
run:
  score: 0.1
"""
        flexmock(StepPrescription).should_receive("_should_include_base").replace_with(lambda _: True).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(StepPrescription.should_include(builder_context)) == [
            {
                "package_name": None,
                "prescription": {
                    "run": False,
                },
                "match": {
                    "package_version": {
                        "index_url": "https://thoth-station.ninja",
                    },
                },
                "multi_package_resolution": False,
                "run": {
                    "score": 0.1,
                },
            }
        ]

    def test_should_include_multi(self) -> None:
        """Test including this pipeline unit multiple times."""
        prescription_str = """
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  - package_version:
      index_url: 'https://thoth-station.ninja'
  - package_version:
      name: flask
run:
  score: 0.1
"""
        flexmock(StepPrescription).should_receive("_should_include_base").replace_with(lambda _: True).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)

        builder_context = flexmock()
        result = list(StepPrescription.should_include(builder_context))
        assert result == [
            {
                "package_name": None,
                "multi_package_resolution": False,
                "prescription": {
                    "run": False,
                },
                "match": {
                    "package_version": {
                        "index_url": "https://thoth-station.ninja",
                    }
                },
                "run": {
                    "score": 0.1,
                },
            },
            {
                "package_name": "flask",
                "multi_package_resolution": False,
                "prescription": {
                    "run": False,
                },
                "match": {
                    "package_version": {
                        "name": "flask",
                    }
                },
                "run": {
                    "score": 0.1,
                },
            },
        ]
        assert result[0]["prescription"] is result[1]["prescription"]

    def test_no_should_include(self) -> None:
        """Test not including this pipeline."""
        prescription_str = """
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    index_url: 'https://thoth-station.ninja'
run:
  score: 0.1
"""
        flexmock(StepPrescription).should_receive("_should_include_base").replace_with(lambda _: False).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(StepPrescription.should_include(builder_context)) == []

    def test_run_state_not_index_url(self, context: Context, state: State) -> None:
        """Test running the prescription if state matches."""
        prescription_str = """
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
    version: '==1.19.1'
    index_url:
      not: 'https://pypi.org/simple'
  state:
    resolved_dependencies:
      - name: tensorflow
        version: '~=2.4.0'
        index_url:
          not: 'https://pypi.org/simple'
run:
  score: 0.5
  stack_info:
  - type: WARNING
    message: Hello, Thoth!
    link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=Source("https://thoth-station.ninja/simple"),
            develop=False,
        )

        state.add_resolved_dependency(("tensorflow", "2.4.0", "https://thoth-station.ninja/simple"))

        unit = StepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = unit.run(state, package_version)
            assert isinstance(result, tuple)
            assert result[0] == 0.5
            assert result[1] is None

            self.verify_justification_schema(context.stack_info)
            assert context.stack_info == [
                {"type": "WARNING", "message": "Hello, Thoth!", "link": "https://thoth-station.ninja"}
            ]

            # Run one more time to make sure stack info is added just once.
            result = unit.run(state, package_version)
            assert isinstance(result, tuple)
            assert result[0] == 0.5
            assert result[1] is None
            assert context.stack_info == [
                {"type": "WARNING", "message": "Hello, Thoth!", "link": "https://thoth-station.ninja"}
            ]

    def test_run_state_not_index_url_not_match(self, context: Context, state: State) -> None:
        """Test running the prescription if state matches."""
        prescription_str = """
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
    version: '==1.19.1'
    index_url:
      not: 'https://pypi.org/simple'
  state:
    resolved_dependencies:
      - name: tensorflow
        version: '~=2.4.0'
        index_url:
          not: 'https://pypi.org/simple'
run:
  score: 0.5
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=Source("https://pypi.org/simple"),  # This does match.
            develop=False,
        )

        state.add_resolved_dependency(("tensorflow", "2.4.0", "https://thoth-station.ninja/simple"))

        unit = StepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state, package_version) is None

    def test_run_state_not_index_url_not_match_state(self, context: Context, state: State) -> None:
        """Test running the prescription if state matches."""
        prescription_str = """
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
    version: '==1.19.1'
    index_url:
      not: 'https://pypi.org/simple'
  state:
    resolved_dependencies:
      - name: tensorflow
        version: '~=2.4.0'
        index_url:
          not: 'https://pypi.org/simple'
run:
  score: 0.5
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=Source("https://thoth-station.ninja/simple"),
            develop=False,
        )

        state.add_resolved_dependency(("tensorflow", "2.4.0", "https://pypi.org/simple"))  # This does match.

        unit = StepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state, package_version) is None

    @pytest.mark.parametrize("develop", [True, False])
    def test_run_develop_match(self, context: Context, state: State, develop: bool) -> None:
        """Test running the prescription if develop flag is set."""
        prescription_str = f"""
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
    develop: {'true' if develop else 'false'}
run:
  score: 0.5
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=Source("https://thoth-station.ninja/simple"),
            develop=develop,
        )

        unit = StepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = unit.run(state, package_version)

        assert isinstance(result, tuple)
        assert result[0] == 0.5
        assert result[1] is None

    @pytest.mark.parametrize("develop", [True, False])
    def test_run_develop_not_match(self, context: Context, state: State, develop: bool) -> None:
        """Test not running the prescription if develop flag is set."""
        prescription_str = f"""
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
    develop: {'true' if develop else 'false'}
run:
  score: 0.5
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=Source("https://thoth-station.ninja/simple"),
            develop=not develop,
        )

        unit = StepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state, package_version) is None

    @pytest.mark.parametrize(
        "develop,state_develop",
        [
            (True, True),
            (False, True),
            (True, False),
            (False, False),
        ],
    )
    def test_run_develop_state_match(self, context: Context, state: State, develop: bool, state_develop: bool) -> None:
        """Test not running the prescription if develop flag is set also on the state match."""
        prescription_str = f"""
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
    develop: {'true' if develop else 'false'}
  state:
    resolved_dependencies:
    - name: pytest
      develop: {'true' if state_develop else 'false'}
run:
  score: 0.5
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=Source("https://pypi.org/simple"),
            develop=develop,
        )

        state_package_version = PackageVersion(
            name="pytest",
            version="==6.2.4",
            index=Source("https://pypi.org/simple"),
            develop=state_develop,
        )
        state.add_resolved_dependency(state_package_version.to_tuple())
        context.register_package_version(state_package_version)

        unit = StepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = unit.run(state, package_version)

        assert isinstance(result, tuple)
        assert result[0] == 0.5
        assert result[1] is None

    @pytest.mark.parametrize(
        "package_version_from_version,package_version_from_index,package_version_from_develop,"
        "register_dependency,pipeline_run",
        [
            ("<=0.25.0", "https://pypi.org/simple", "false", True, True),
            (">0.25.0", "https://pypi.org/simple", "false", True, False),  # Version not matching.
            ("<=0.25.0", '{"not": "https://pypi.org/simple"}', "false", True, False),  # Index not matching.
            ("<=0.25.0", "https://pypi.org/simple", "true", True, False),  # Develop flag not matching.
            ("<=0.25.0", "https://pypi.org/simple", "false", False, False),  # Dependency is not registered.
        ],
    )
    def test_run_package_version_from(
        self,
        context: Context,
        state: State,
        package_version_from_version: str,
        package_version_from_index: str,
        package_version_from_develop: str,
        register_dependency: bool,
        pipeline_run: bool,
    ) -> None:
        """Test running the prescription based on the dependency introduced."""
        prescription_str = f"""
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
  state:
    package_version_from:
    - name: scikit-learn
      version: "{package_version_from_version}"
      develop: {package_version_from_develop}
      index_url: {package_version_from_index}
run:
  score: 0.5
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)

        pypi = Source("https://pypi.org/simple")
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=pypi,
            develop=False,
        )

        package_version_from = PackageVersion(
            name="scikit-learn",
            version="==0.24.2",
            index=pypi,
            develop=False,
        )

        state.add_resolved_dependency(package_version_from.to_tuple())
        context.register_package_version(package_version_from)

        if register_dependency:
            runtime_env = context.project.runtime_environment
            context.register_package_tuple(
                package_version.to_tuple(),
                dependent_tuple=package_version_from.to_tuple(),
                develop=False,
                extras=None,
                os_name=runtime_env.operating_system.name,
                os_version=runtime_env.operating_system.version,
                python_version=runtime_env.python_version,
            )

        unit = StepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = unit.run(state, package_version)

        if pipeline_run:
            assert isinstance(result, tuple)
            assert result[0] == 0.5
            assert result[1] is None
        else:
            assert result is None

    @pytest.mark.parametrize(
        "package_version_from_version,package_version_from_index,package_version_from_develop,"
        "resolved_version,resolved_index,resolved_develop,"
        "add_resolved,pipeline_run",
        [
            ("<=0.25.0", "https://pypi.org/simple", "false", "==8.0.0", "https://pypi.org/simple", "false", True, True),
            # Version not matching.
            (">0.25.0", "https://pypi.org/simple", "false", "==8.0.0", "https://pypi.org/simple", "false", True, False),
            # Index not matching.
            (
                "<=0.25.0",
                '{"not": "https://pypi.org/simple"}',
                "false",
                "==8.0.0",
                "https://pypi.org/simple",
                "false",
                True,
                False,
            ),
            # Develop flag not matching.
            ("<=0.25.0", "https://pypi.org/simple", "true", "==8.0.0", "https://pypi.org/simple", "false", True, False),
            # Version not matching.
            (
                "<=0.25.0",
                "https://pypi.org/simple",
                "false",
                "!=8.0.0",
                "https://pypi.org/simple",
                "false",
                True,
                False,
            ),
            # Index not matching.
            (
                "<=0.25.0",
                "https://pypi.org/simple",
                "false",
                "==8.0.0",
                '{"not": "https://pypi.org/simple"}',
                "false",
                True,
                False,
            ),
            # Develop flag not matching.
            ("<=0.25.0", "https://pypi.org/simple", "false", "==8.0.0", "https://pypi.org/simple", "true", True, False),
            # Resolved dependency nit present.
            (
                "<=0.25.0",
                "https://pypi.org/simple",
                "false",
                "==8.0.0",
                "https://pypi.org/simple",
                "false",
                False,
                False,
            ),
        ],
    )
    def test_run_package_version_from_with_resolved(
        self,
        context: Context,
        state: State,
        package_version_from_version: str,
        package_version_from_index: str,
        package_version_from_develop: str,
        resolved_version: str,
        resolved_index: str,
        resolved_develop: str,
        add_resolved: bool,
        pipeline_run: bool,
    ) -> None:
        """Test running the prescription based on the dependency introduced."""
        prescription_str = f"""
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
  state:
    package_version_from:
    - name: scikit-learn
      version: "{package_version_from_version}"
      develop: {package_version_from_develop}
      index_url: {package_version_from_index}
    resolved_dependencies:
    - name: click
      version: "{resolved_version}"
      develop: {resolved_develop}
      index_url: {resolved_index}
run:
  score: 0.5
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)

        pypi = Source("https://pypi.org/simple")
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=pypi,
            develop=False,
        )

        package_version_from = PackageVersion(
            name="scikit-learn",
            version="==0.24.2",
            index=pypi,
            develop=False,
        )

        package_version_resolved = PackageVersion(
            name="click",
            version="==8.0.0",
            index=pypi,
            develop=False,
        )

        state.add_resolved_dependency(package_version_from.to_tuple())
        context.register_package_version(package_version_from)

        if add_resolved:
            state.add_resolved_dependency(package_version_resolved.to_tuple())
            context.register_package_version(package_version_resolved)

        runtime_env = context.project.runtime_environment
        context.register_package_tuple(
            package_version.to_tuple(),
            dependent_tuple=package_version_from.to_tuple(),
            develop=False,
            extras=None,
            os_name=runtime_env.operating_system.name,
            os_version=runtime_env.operating_system.version,
            python_version=runtime_env.python_version,
        )

        unit = StepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = unit.run(state, package_version)

        if pipeline_run:
            assert isinstance(result, tuple)
            assert result[0] == 0.5
            assert result[1] is None
        else:
            assert result is None

    @pytest.mark.parametrize("allow_other", [True, False])
    def test_run_package_version_from_with_other(self, context: Context, state: State, allow_other: bool) -> None:
        """Test running the prescription based on the dependency introduced without with considering other packages."""
        prescription_str = f"""
name: StepUnit
type: step
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
  state:
    package_version_from_allow_other: {'true' if allow_other else 'false'}
    package_version_from:
    - name: scikit-learn
      version: "<1.0.0"
      develop: false
      index_url: https://pypi.org/simple
run:
  score: 0.5
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_STEP_SCHEMA(prescription)
        StepPrescription.set_prescription(prescription)

        pypi = Source("https://pypi.org/simple")

        package_version = PackageVersion(
            name="numpy",
            version="==1.19.4",
            index=pypi,
            develop=False,
        )

        package_version_other = PackageVersion(
            name="tensorflow",
            version="==2.6.0",
            index=pypi,
            develop=False,
        )

        package_version_from = PackageVersion(
            name="scikit-learn",
            version="==0.24.2",
            index=pypi,
            develop=False,
        )

        state.add_resolved_dependency(package_version_from.to_tuple())
        context.register_package_version(package_version_from)

        state.add_resolved_dependency(package_version_other.to_tuple())
        context.register_package_version(package_version_other)

        runtime_env = context.project.runtime_environment
        context.register_package_tuple(
            package_version.to_tuple(),
            dependent_tuple=package_version_from.to_tuple(),
            develop=False,
            extras=None,
            os_name=runtime_env.operating_system.name,
            os_version=runtime_env.operating_system.version,
            python_version=runtime_env.python_version,
        )

        context.register_package_tuple(
            package_version.to_tuple(),
            dependent_tuple=package_version_other.to_tuple(),
            develop=False,
            extras=None,
            os_name=runtime_env.operating_system.name,
            os_version=runtime_env.operating_system.version,
            python_version=runtime_env.python_version,
        )

        unit = StepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = unit.run(state, package_version)

        if allow_other:
            assert isinstance(result, tuple)
            assert result[0] == 0.5
            assert result[1] is None
        else:
            assert result is None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation of pipeline builder - create pipeline configuration based on the project and its library usage."""

import os
import logging
import json
from typing import Any
from typing import Dict
from typing import Generator
from typing import Type
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from itertools import chain

import attr
from thoth.python import Project
from thoth.storages import GraphDatabase
import yaml

from .enums import DecisionType
from .enums import RecommendationType
from .exceptions import InternalError
from .exceptions import UnknownPipelineUnitError
from .exceptions import PipelineConfigurationError
from .pipeline_config import PipelineConfig
from .prescription import UnitPrescription

if TYPE_CHECKING:
    from .prescription import Prescription  # noqa: F401
    from .unit_types import BootType
    from .unit_types import PseudonymType
    from .unit_types import SieveType
    from .unit_types import StepType
    from .unit_types import StrideType
    from .unit_types import UnitType
    from .unit_types import WrapType

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class PipelineBuilderContext:
    """A context passed to units to determine if they want to participate in a pipeline."""

    graph = attr.ib(type=GraphDatabase, kw_only=True, default=None)
    project = attr.ib(type=Project, kw_only=True, default=None)
    library_usage = attr.ib(type=Optional[Dict[str, Any]], kw_only=True, default=None)
    labels = attr.ib(type=Dict[str, str], kw_only=True, default=attr.Factory(dict))
    decision_type = attr.ib(type=Optional[DecisionType], kw_only=True, default=None)
    recommendation_type = attr.ib(type=Optional[RecommendationType], kw_only=True, default=None)
    cli_parameters = attr.ib(type=Dict[str, Any], kw_only=True, default=attr.Factory(dict))
    prescription = attr.ib(type=Optional["Prescription"], kw_only=True, default=None)
    iteration = attr.ib(type=int, kw_only=True, default=0)
    authenticated = attr.ib(type=bool, kw_only=True)

    _boots = attr.ib(type=Dict[Optional[str], List["BootType"]], factory=dict, kw_only=True)
    _pseudonyms = attr.ib(type=Dict[str, List["PseudonymType"]], factory=dict, kw_only=True)
    _sieves = attr.ib(type=Dict[Optional[str], List["SieveType"]], factory=dict, kw_only=True)
    _steps = attr.ib(type=Dict[Optional[str], List["StepType"]], factory=dict, kw_only=True)
    _strides = attr.ib(type=Dict[Optional[str], List["StrideType"]], factory=dict, kw_only=True)
    _wraps = attr.ib(type=Dict[Optional[str], List["WrapType"]], factory=dict, kw_only=True)
    _boots_included = attr.ib(type=Dict[str, List["BootType"]], factory=dict, kw_only=True)
    _pseudonyms_included = attr.ib(type=Dict[str, List["PseudonymType"]], factory=dict, kw_only=True)
    _sieves_included = attr.ib(type=Dict[str, List["SieveType"]], factory=dict, kw_only=True)
    _steps_included = attr.ib(type=Dict[str, List["StepType"]], factory=dict, kw_only=True)
    _strides_included = attr.ib(type=Dict[str, List["StrideType"]], factory=dict, kw_only=True)
    _wraps_included = attr.ib(type=Dict[str, List["WrapType"]], factory=dict, kw_only=True)

    @authenticated.default
    def _authenticated_default(self) -> bool:
        """Check if adviser is running in an authenticated mode."""
        env_authenticated = os.getenv("THOTH_AUTHENTICATED_ADVISE")
        return bool(int(env_authenticated)) if env_authenticated is not None else False

    @property
    def boots(self) -> List["BootType"]:
        """Get all boots registered to this pipeline builder context."""
        return list(chain(*self._boots.values()))

    @property
    def boots_dict(self) -> Dict[Optional[str], List["BootType"]]:
        """Get boots as a dictionary mapping."""
        return self._boots

    @property
    def pseudonyms(self) -> List["PseudonymType"]:
        """Get all pseudonyms registered to this pipeline builder context."""
        return list(chain(*self._pseudonyms.values()))

    @property
    def pseudonyms_dict(self) -> Dict[str, List["PseudonymType"]]:
        """Get pseudonyms as a dictionary mapping."""
        return self._pseudonyms

    @property
    def sieves(self) -> List["SieveType"]:
        """Get all sieves registered to this pipeline builder context."""
        return list(chain(*self._sieves.values()))

    @property
    def sieves_dict(self) -> Dict[Optional[str], List["SieveType"]]:
        """Get sieves as a dictionary mapping."""
        return self._sieves

    @property
    def steps(self) -> List["StepType"]:
        """Get all steps registered to this pipeline builder context."""
        return list(chain(*self._steps.values()))

    @property
    def steps_dict(self) -> Dict[Optional[str], List["StepType"]]:
        """Get steps as a dictionary mapping."""
        return self._steps

    @property
    def strides(self) -> List["StrideType"]:
        """Get all strides registered to this pipeline builder context."""
        return list(chain(*self._strides.values()))

    @property
    def strides_dict(self) -> Dict[Optional[str], List["StrideType"]]:
        """Get strides as a dictionary mapping."""
        return self._strides

    @property
    def wraps(self) -> List["WrapType"]:
        """Get all wraps registered to this pipeline builder context."""
        return list(chain(*self._wraps.values()))

    @property
    def wraps_dict(self) -> Dict[Optional[str], List["WrapType"]]:
        """Get wraps as a dictionary mapping."""
        return self._wraps

    def __attrs_post_init__(self) -> None:
        """Verify we have only adviser or dependency monkey specific builder."""
        if self.decision_type is not None and self.recommendation_type is not None:
            raise ValueError("Cannot instantiate builder for adviser and dependency monkey at the same time")

        if self.decision_type is None and self.recommendation_type is None:
            raise ValueError("Cannot instantiate builder context not specific to adviser nor dependency monkey")

    def is_included(self, unit_class: Type["UnitType"]) -> bool:
        """Check if the given pipeline unit is already included in the pipeline configuration."""
        if unit_class.is_boot_unit_type():
            return unit_class.get_unit_name() in self._boots_included
        elif unit_class.is_pseudonym_unit_type():
            return unit_class.get_unit_name() in self._pseudonyms_included
        elif unit_class.is_sieve_unit_type():
            return unit_class.get_unit_name() in self._sieves_included
        elif unit_class.is_step_unit_type():
            return unit_class.get_unit_name() in self._steps_included
        elif unit_class.is_stride_unit_type():
            return unit_class.get_unit_name() in self._strides_included
        elif unit_class.is_wrap_unit_type():
            return unit_class.get_unit_name() in self._wraps_included

        raise InternalError(f"Unknown unit {unit_class.get_unit_name()!r} of type {unit_class}")

    def get_included_boots(self, boot_class: Type["UnitType"]) -> Generator["BootType", None, None]:
        """Get included boots of the provided boot class."""
        assert boot_class.is_boot_unit_type()
        yield from self._boots_included.get(boot_class.get_unit_name(), [])

    def get_included_boot_names(self) -> Generator[str, None, None]:
        """Get names of included boots."""
        yield from self._boots_included.keys()

    def get_included_pseudonyms(self, pseudonym_class: Type["PseudonymType"]) -> Generator["PseudonymType", None, None]:
        """Get included sieves of the provided sieve class."""
        assert pseudonym_class.is_pseudonym_unit_type()
        yield from self._pseudonyms_included.get(pseudonym_class.get_unit_name(), [])

    def get_included_pseudonym_names(self) -> Generator[str, None, None]:
        """Get names of included pseudonyms."""
        yield from self._pseudonyms_included.keys()

    def get_included_sieves(self, sieve_class: Type["SieveType"]) -> Generator["SieveType", None, None]:
        """Get included sieves of the provided sieve class."""
        assert sieve_class.is_sieve_unit_type()
        yield from self._sieves_included.get(sieve_class.get_unit_name(), [])

    def get_included_sieve_names(self) -> Generator[str, None, None]:
        """Get names of included sieves."""
        yield from self._sieves_included.keys()

    def get_included_steps(self, step_class: Type["StepType"]) -> Generator["StepType", None, None]:
        """Get included steps of the provided step class."""
        assert step_class.is_step_unit_type()
        yield from self._steps_included.get(step_class.get_unit_name(), [])

    def get_included_step_names(self) -> Generator[str, None, None]:
        """Get names of included steps."""
        yield from self._steps_included.keys()

    def get_included_strides(self, stride_class: Type["StrideType"]) -> Generator["StrideType", None, None]:
        """Get included strides of the provided stride class."""
        assert stride_class.is_stride_unit_type()
        yield from self._strides_included.get(stride_class.get_unit_name(), [])

    def get_included_stride_names(self) -> Generator[str, None, None]:
        """Get names of included strides."""
        yield from self._strides_included.keys()

    def get_included_wraps(self, wrap_class: Type["WrapType"]) -> Generator["WrapType", None, None]:
        """Get included wraps of the provided wrap class."""
        assert wrap_class.is_wrap_unit_type()
        yield from self._wraps_included.get(wrap_class.get_unit_name(), [])

    def get_included_wrap_names(self) -> Generator[str, None, None]:
        """Get names of included wraps."""
        yield from self._wraps_included.keys()

    def is_adviser_pipeline(self) -> bool:
        """Check if the pipeline built is meant for adviser."""
        return self.decision_type is None and self.recommendation_type is not None

    def is_dependency_monkey_pipeline(self) -> bool:
        """Check if the pipeline built is meant for Dependency Monkey."""
        return self.decision_type is not None and self.recommendation_type is None

    def add_unit(self, unit: "UnitType") -> None:
        """Add the given unit to pipeline configuration."""
        package_name: Optional[str] = unit.configuration.get("package_name")

        if unit.is_boot_unit_type():
            self._boots_included.setdefault(unit.name, []).append(unit)
            self._boots.setdefault(package_name, []).append(unit)
            return
        elif unit.is_pseudonym_unit_type():
            if not package_name:
                raise PipelineConfigurationError(
                    f"Pipeline cannot be constructed as unit {unit.name!r} of type Pseudonym "
                    f"did not provide any package name configuration: {unit.configuration!r}"
                )
            self._pseudonyms_included.setdefault(unit.name, []).append(unit)
            self._pseudonyms.setdefault(package_name, []).append(unit)
            return
        elif unit.is_sieve_unit_type():
            self._sieves_included.setdefault(unit.name, []).append(unit)
            self._sieves.setdefault(package_name, []).append(unit)
            return
        elif unit.is_step_unit_type():
            self._steps_included.setdefault(unit.name, []).append(unit)
            self._steps.setdefault(package_name, []).append(unit)
            return
        elif unit.is_stride_unit_type():
            self._strides_included.setdefault(unit.name, []).append(unit)
            self._strides.setdefault(package_name, []).append(unit)
            return
        elif unit.is_wrap_unit_type():
            self._wraps_included.setdefault(unit.name, []).append(unit)
            self._wraps.setdefault(package_name, []).append(unit)
            return

        raise InternalError(f"Unknown unit {unit!r} of type {unit.name!r}")


class PipelineBuilder:
    """Builder responsible for creating pipeline configuration from the project and its library usage."""

    __slots__: List[str] = []

    def __init__(self) -> None:
        """Instantiate of the pipeline builder - do NOT instantiate this class."""
        raise NotImplementedError("Cannot instantiate pipeline builder")

    @staticmethod
    def _iter_units(ctx: PipelineBuilderContext) -> Generator["UnitType", None, None]:
        """Iterate over pipeline units available in this implementation."""
        # Imports placed here to simplify tests.
        import thoth.adviser.boots
        import thoth.adviser.pseudonyms
        import thoth.adviser.sieves
        import thoth.adviser.steps
        import thoth.adviser.strides
        import thoth.adviser.wraps

        for boot_name in thoth.adviser.boots.__all__:
            yield getattr(thoth.adviser.boots, boot_name)
        if ctx.prescription:
            yield from ctx.prescription.iter_boot_units()

        for pseudonym_name in thoth.adviser.pseudonyms.__all__:
            yield getattr(thoth.adviser.pseudonyms, pseudonym_name)
        if ctx.prescription:
            yield from ctx.prescription.iter_pseudonym_units()

        for sieve_name in thoth.adviser.sieves.__all__:
            yield getattr(thoth.adviser.sieves, sieve_name)
        if ctx.prescription:
            yield from ctx.prescription.iter_sieve_units()

        for step_name in thoth.adviser.steps.__all__:
            yield getattr(thoth.adviser.steps, step_name)
        if ctx.prescription:
            yield from ctx.prescription.iter_step_units()

        for stride_name in thoth.adviser.strides.__all__:
            yield getattr(thoth.adviser.strides, stride_name)
        if ctx.prescription:
            yield from ctx.prescription.iter_stride_units()

        for wrap_name in thoth.adviser.wraps.__all__:
            yield getattr(thoth.adviser.wraps, wrap_name)
        if ctx.prescription:
            yield from ctx.prescription.iter_wrap_units()

    @classmethod
    def _build_configuration(cls, ctx: PipelineBuilderContext) -> PipelineConfig:
        """Instantiate units and return the actual pipeline configuration."""
        # As pipeline steps can have dependencies on each other, iterate over them until we have any change done
        # to the pipeline configuration.
        _LOGGER.info("Creating pipeline configuration")

        blocked_units = (
            set(os.environ["THOTH_ADVISER_BLOCKED_UNITS"].split(","))
            if "THOTH_ADVISER_BLOCKED_UNITS" in os.environ
            else set()
        )

        change = True
        ctx.iteration = -1
        try:
            while change:
                change = False
                ctx.iteration += 1
                for unit_class in cls._iter_units(ctx):
                    unit_name = unit_class.get_unit_name()
                    if unit_name in blocked_units:
                        _LOGGER.debug(
                            "Avoiding adding pipeline unit %r based on blocked units configuration",
                            unit_name,
                        )
                        continue

                    for unit_configuration in unit_class.should_include(ctx):
                        if unit_configuration is None:
                            _LOGGER.debug(
                                "Pipeline unit %r will not be included in the pipeline configuration in this round",
                                unit_name,
                            )
                            continue

                        change = True

                        _LOGGER.debug(
                            "Including pipeline unit %r in pipeline configuration with unit configuration %r",
                            unit_name,
                            unit_configuration,
                        )
                        unit_instance = unit_class()  # type: ignore

                        # Always perform update, even with an empty dict. Update triggers a schema check.
                        try:
                            unit_instance.update_configuration(unit_configuration)
                        except Exception as exc:
                            raise PipelineConfigurationError(
                                f"Filed to initialize pipeline unit configuration for {unit_name!r} "
                                f"with configuration {unit_configuration!r}: {str(exc)}"
                            ) from exc

                        ctx.add_unit(unit_instance)
        finally:
            # Once the build pipeline is constructed or fails to construct, we can clear cached results.
            UnitPrescription.SHOULD_INCLUDE_CACHE.clear()

        pipeline = PipelineConfig(
            boots=ctx.boots_dict,
            pseudonyms=ctx.pseudonyms_dict,
            sieves=ctx.sieves_dict,
            steps=ctx.steps_dict,
            strides=ctx.strides_dict,
            wraps=ctx.wraps_dict,
        )

        if _LOGGER.getEffectiveLevel() <= logging.DEBUG:
            _LOGGER.debug(
                "Pipeline configuration creation ended, configuration:\n%s",
                json.dumps(pipeline.to_dict(), indent=2),
            )

        return pipeline

    @staticmethod
    def _do_instantiate_from_dict(module: object, configuration_entry: Dict[str, Any]) -> "UnitType":
        """Instantiate a pipeline unit from a dict representation."""
        if "name" not in configuration_entry:
            raise ValueError(f"No pipeline unit name provided in the configuration entry: {configuration_entry!r}")

        try:
            unit_class = getattr(module, configuration_entry["name"])
        except AttributeError as exc:
            raise UnknownPipelineUnitError(f"Cannot import unit {configuration_entry['name']}: {str(exc)}") from exc

        unit: "UnitType" = unit_class()

        if configuration_entry.get("configuration"):
            try:
                unit.update_configuration(configuration_entry["configuration"])
            except Exception as exc:
                raise PipelineConfigurationError(
                    f"Filed to initialize pipeline unit configuration for {unit.name!r} "
                    f"with configuration {configuration_entry['configuration']!r}: {str(exc)}"
                ) from exc

        return unit

    @classmethod
    def from_dict(cls, dict_: Dict[str, Any]) -> "PipelineConfig":
        """Instantiate pipeline configuration based on dictionary supplied."""
        # Imports placed here to simplify tests.
        import thoth.adviser.boots
        import thoth.adviser.pseudonyms
        import thoth.adviser.sieves
        import thoth.adviser.steps
        import thoth.adviser.strides
        import thoth.adviser.wraps

        boots: Dict[Optional[str], List["BootType"]] = {}
        for boot_entry in dict_.pop("boots", []) or []:
            boot_unit: "BootType" = cls._do_instantiate_from_dict(thoth.adviser.boots, boot_entry)
            package_name = boot_unit.configuration.get("package_name")
            boots.setdefault(package_name, []).append(boot_unit)

        pseudonyms: Dict[str, List["PseudonymType"]] = {}
        for pseudonym_entry in dict_.pop("pseudonyms", []) or []:
            unit: "PseudonymType" = cls._do_instantiate_from_dict(thoth.adviser.pseudonyms, pseudonym_entry)

            package_name = unit.configuration.get("package_name")
            if not package_name:
                raise PipelineConfigurationError(
                    f"Pipeline cannot be constructed as unit {unit.name!r} of type Pseudonym "
                    f"did not provide any package name configuration: {unit.configuration!r}"
                )

            pseudonyms.setdefault(package_name, []).append(unit)

        sieves: Dict[Optional[str], List["SieveType"]] = {}
        for sieve_entry in dict_.pop("sieves", []) or []:
            sieve_unit: "SieveType" = cls._do_instantiate_from_dict(thoth.adviser.sieves, sieve_entry)
            package_name = sieve_unit.configuration.get("package_name")
            sieves.setdefault(package_name, []).append(sieve_unit)

        steps: Dict[Optional[str], List["StepType"]] = {}
        for step_entry in dict_.pop("steps", []) or []:
            step_unit: "StepType" = cls._do_instantiate_from_dict(thoth.adviser.steps, step_entry)
            package_name = step_unit.configuration.get("package_name")
            steps.setdefault(package_name, []).append(step_unit)

        strides: Dict[Optional[str], List["StrideType"]] = {}
        for stride_entry in dict_.pop("strides", []) or []:
            stride_unit: "StrideType" = cls._do_instantiate_from_dict(thoth.adviser.strides, stride_entry)
            package_name = stride_unit.configuration.get("package_name")
            strides.setdefault(package_name, []).append(stride_unit)

        wraps: Dict[Optional[str], List["WrapType"]] = {}
        for wrap_entry in dict_.pop("wraps", []) or []:
            wrap_unit: "WrapType" = cls._do_instantiate_from_dict(thoth.adviser.wraps, wrap_entry)
            package_name = wrap_unit.configuration.get("package_name")
            wraps.setdefault(package_name, []).append(wrap_unit)

        if dict_:
            _LOGGER.warning("Unknown entry in pipeline configuration: %r", dict_)

        pipeline = PipelineConfig(
            boots=boots,
            pseudonyms=pseudonyms,
            sieves=sieves,
            steps=steps,
            strides=strides,
            wraps=wraps,
        )
        _LOGGER.debug(
            "Pipeline configuration creation ended, configuration:\n%s",
            json.dumps(pipeline.to_dict(), indent=2),
        )
        return pipeline

    @classmethod
    def load(cls, config: str) -> "PipelineConfig":
        """Load pipeline configuration from a file or a string."""
        if os.path.isfile(config):
            _LOGGER.debug("Loading pipeline configuration from file %r", config)
            with open(config, "r") as config_file:
                config = config_file.read()

        return cls.from_dict(yaml.safe_load(config))

    @classmethod
    def get_adviser_pipeline_config(
        cls,
        *,
        recommendation_type: RecommendationType,
        graph: GraphDatabase,
        project: Project,
        labels: Dict[str, str],
        library_usage: Optional[Dict[str, Any]],
        prescription: Optional["Prescription"],
        cli_parameters: Dict[str, Any],
    ) -> PipelineConfig:
        """Get adviser's pipeline configuration."""
        return cls._build_configuration(
            PipelineBuilderContext(
                graph=graph,
                project=project,
                labels=labels,
                library_usage=library_usage,
                recommendation_type=recommendation_type,
                prescription=prescription,
                cli_parameters=cli_parameters,
            )
        )

    @classmethod
    def get_dependency_monkey_pipeline_config(
        cls,
        *,
        decision_type: DecisionType,
        graph: GraphDatabase,
        project: Project,
        labels: Dict[str, str],
        library_usage: Optional[Dict[str, Any]],
        prescription: Optional["Prescription"],
        cli_parameters: Dict[str, Any],
    ) -> PipelineConfig:
        """Get dependency-monkey's pipeline configuration."""
        return cls._build_configuration(
            PipelineBuilderContext(
                graph=graph,
                project=project,
                labels=labels,
                library_usage=library_usage,
                decision_type=decision_type,
                prescription=prescription,
                cli_parameters=cli_parameters,
            )
        )
